#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fabx_parser.h"
#include "fabx_rpn.h"
#include "fabn.h"
#include "fabn_rpneval.h"
#include "dyn_string.h"

int main(int argc, char **argv) {
	Dynamic_str buffer;
	int no_of_comm, forced_no_of_comm=-1, contains_if_iff;
	char *pom;
	FABN result;
	FABN *values;

	if (argc<2) {
		fprintf(stderr, "Usage: %s \"<formula>\" [<no_of_comm>]\n", argv[0]);
		return 1;
	}
	if (argc==3) {
		forced_no_of_comm = atoi(argv[2]);
		if (forced_no_of_comm<0 || forced_no_of_comm>9) {
			fprintf(stderr, "Error: Invalid number of generators. Only values 0-9 are allowed.\n");
			return 1;
		}	
	}

	init_string(&buffer);

	if (!to_RPN_ver5_autodetect(argv[1], &no_of_comm, &buffer)) {
		if (forced_no_of_comm>=0 && forced_no_of_comm<no_of_comm) {
			fprintf(stderr, "Error: Number of desired generators (%d) is too low.\nIt should be at least %d.\n", forced_no_of_comm, no_of_comm);
			release_string(&buffer);
			return -1;
		}
		if (forced_no_of_comm>no_of_comm) no_of_comm = forced_no_of_comm;
		buffer.str[0] = '0'+no_of_comm;
		printf("No.of.comm: %c\n", buffer.str[0]);
		printf("Expr: %s\n", argv[1]);
		printf("RPN: %s\n", &buffer.str[1]);	

		/* check whether contains if or iff */
		contains_if_iff=0;
		pom=buffer.str;
		while (*pom && *(pom+1)) {
			if (*pom=='i' && *(pom+1)=='f') {
				contains_if_iff=1;
				break;
			}
			pom++;
		}

		contains_if_iff = 1-contains_if_iff;

		create_fabn(&result, no_of_comm);

		if (create_std_values_fabn(no_of_comm, &values)<0) {
			release_string(&buffer);
			F_fabn(&result);
			return -1;
		}
		eval_RPN_in_fabn_ver3(buffer.str,&values,&result, contains_if_iff);

		clear_std_values_fabn(no_of_comm, &values);

		ptb_fabn(&result);
		clear_fabn(&result);
	}
	
	release_string(&buffer);

	return 0;
}

