#ifndef _FREE_ORTHOMODULAR_AB_
#define _FREE_ORTHOMODULAR_AB_

typedef unsigned char FAB; 

extern const FAB A, B, cA, cB, O, I, T, F;

int is_bool_fab(FAB f);
FAB addapt_bool_fab(int bool);
void normalize_FAB(FAB *f);
FAB complement_FAB(FAB *f);
FAB join_FAB(FAB *f1, FAB *f2);
FAB meet_FAB(FAB *f1, FAB *f2);
void tex_FAB(FAB *f); 
void print_FAB(FAB *f);

/* Functions print n-th row of text-graphical representation of the element */
void prrow1st_fab(FAB f);
void prrow2nd_fab(FAB f);
void prrow3rd_fab(FAB f);
void prrow4th_fab(FAB f);
void prrow5th_fab(FAB f);

/* Nullary operations - constants */
/* Zero */
FAB O_fab(void);
/* Element a */
FAB A_fab(void);
/* Element b */
FAB B_fab(void);
/* Unit */
FAB I_fab(void);

/* Complement */
FAB c_fab(FAB f);

/* Join */
FAB j_fab(FAB f1, FAB f2);

/* Meet */
FAB m_fab(FAB f1, FAB f2);

/* Skew join (f1/\f2')\/f2 */
FAB sj_fab(FAB f1, FAB f2);

/* Skew meet (f1\/f2')/\f2 */
FAB sm_fab(FAB f1, FAB f2);

/* Sasaki projection \varphi_f1(f2) = f1/\(f1'\/f2) */
FAB sp_fab(FAB f1, FAB f2);

/* Implications and Equivalence from : 
	 - Quantumm and Classical Implication Algebras with Primitive Implications
	 - Equivalence, Identities, Symmetric Differences and Congruences in Orthomodular Lattices
*/

/* classical */

/* f1'\/f2 */
FAB imp0_fab(FAB f1, FAB f2); 

/* (f1'\/f2)/\(f1\/f2')*/
FAB eq0_fab(FAB f1, FAB f2);

/* Sasaki */

/* f1'\/(f1/\f2) */
FAB imp1_fab(FAB f1, FAB f2); 

/* (f1\/f2')/\(f1'\/(f1/\f2)) */
FAB eq1_fab(FAB f1, FAB f2);

/* Dishkant */

/* f2\/(f1'\/f2') */
FAB imp2_fab(FAB f1, FAB f2);

/* (f1\/f2')/\(f2\/(f1'/\f2')) */
FAB eq2_fab(FAB f1, FAB f2);

/* Kalmbach */

/* ((f1'/\f2)\/(f1'/\f2')\/(f1/\(f1'\/f2)) */
FAB imp3_fab(FAB f1, FAB f2); 

/* (f1'\/f2)/\(f1\/(f1'/\f2')) */
FAB eq3_fab(FAB f1, FAB f2);

/* non-tollens */

/* ((f1/\f2)\/(f1'/\f2))\/((f1'\/f2)/\f2') */
FAB imp4_fab(FAB f1, FAB f2); 

/* (f1'\/f2)/\(f1'\/(f1/\f2)) */
FAB eq4_fab(FAB f1, FAB f2);

/* relevance */

/* ((f1/\f2)\/(f1'/\f2))\/(f1'/\f2') */
FAB imp5_fab(FAB f1, FAB f2); 

/* (f1/\f2)\/(f1'/\f2') */
FAB eq5_fab(FAB f1, FAB f2);

/* Lower commutator */
FAB lc_fab(FAB f1, FAB f2); 

/* Upper commutator */
FAB uc_fab(FAB f1, FAB f2); 

/* prints text_graphical interpretation of an element */
/* n-th row of this text_graphical interpretation n=1,2,3,4,5 */
void prnthrow_fab(FAB f, int n);
 
/* prints text_graphical interpretation of an element */
void pr_fab(FAB f);

/* prints TeX macro representation of an element(form2.sty) */
void tex_fab(FAB f);

/* prints Beran's number of an element */
void beran_fab(FAB f);

/* prints both text and TeX representation of an element */
void pt_fab(FAB f);

/* prints text ,TeX representation of an element and its Beran's number*/
void ptb_fab(FAB f);

/* Commutativity relation: prints f1 and (f1/\f2) \/ (f1/\f2') */
void C_pr_fab(FAB f1, FAB f2);
/* returns second argument (rs = right side)*/
FAB C_rs_fab(FAB f1, FAB f2);

/* Returns 1 if f1 C f2 otherwise returns 0 */
int C_i_fab(FAB f1, FAB f2); 

/* Returns FAB boolean values from C_i*/
FAB C_f_fab(FAB f1, FAB f2);

/* Dual Commutativity relation(from Beran): prints f1 and (f1\/f2)/\(f1\/f2') */
void D_pr_fab(FAB f1, FAB f2);
/* returns second argument (rs = right side)*/
FAB D_rs_fab(FAB f1, FAB f2);

/* Returns 1 if f1 C f2 otherwise returns 0 */
int D_i_fab(FAB f1, FAB f2); 

/* Returns FAB boolean values D_i*/
FAB D_f_fab(FAB f1, FAB f2);

/* Inequality relations */

/* Equality */
int is_eq_fab(FAB f1, FAB f2); 

FAB is_eq_f_fab(FAB f1, FAB f2); 

int is_eq_inter_fab(FAB f1, FAB f2, FAB interpretation); 

FAB is_eq_inter_f_fab(FAB f1, FAB f2, FAB interpretation); 

/* less than or equal */
int is_leq_fab(FAB f1, FAB f2); 

FAB is_leq_f_fab(FAB f1, FAB f2); 

int is_leq_inter_fab(FAB f1, FAB f2, FAB interpretation);

FAB is_leq_inter_f_fab(FAB f1, FAB f2, FAB interpretation);

#if 0
/* less than */
int is_lt_fab(FAB f1, FAB f2);

FAB is_lt_f_fab(FAB f1, FAB f2); 

int is_lt_inter_fab(FAB f1, FAB f2, FAB interpretation);

FAB is_lt_inter_f_fab(FAB f1, FAB f2, FAB interpretation);
#endif

/* greater than or equal */
int is_geq_fab(FAB f1, FAB f2);

FAB is_geq_f_fab(FAB f1, FAB f2); 

int is_geq_inter_fab(FAB f1, FAB f2, FAB interpretation);

FAB is_geq_inter_f_fab(FAB f1, FAB f2, FAB interpretation);


#if 0
/* greater than */
int is_gt_fab(FAB f1, FAB f2); 

FAB is_gt_f_fab(FAB f1, FAB f2); 

int is_gt_inter_fab(FAB f1, FAB f2, FAB interpretation);

FAB is_gt_inter_f_fab(FAB f1, FAB f2, FAB interpretation);
#endif

/* Interpretations */

void create_interpretation_eq_fab(FAB f1, FAB f2, FAB *interpretation);

void create_interpretation_leq_fab(FAB f1, FAB f2, FAB *interpretation);

void create_interpretation_geq_fab(FAB f1, FAB f2, FAB *interpretation);

void create_interpretation_C_fab(FAB f1, FAB f2, FAB *interpretation); 

void create_interpretation_D_fab(FAB f1, FAB f2, FAB *interpretation); 

/* Conditional operations */

FAB cond_eq_fab(FAB f1, FAB f2); 
FAB cond_eq_fab_rel(FAB f1, FAB f2, int eval_as_rel); 
FAB cond_leq_fab(FAB f1, FAB f2); 
FAB cond_leq_fab_rel(FAB f1, FAB f2, int eval_as_rel); 
FAB cond_geq_fab(FAB f1, FAB f2); 
FAB cond_geq_fab_rel(FAB f1, FAB f2, int eval_as_rel); 
FAB cond_C_fab(FAB f1, FAB f2); 
FAB cond_C_fab_rel(FAB f1, FAB f2, int eval_as_rel); 
FAB cond_D_fab(FAB f1, FAB f2); 
FAB cond_D_fab_rel(FAB f1, FAB f2, int eval_as_rel); 
FAB cond_and_fab(FAB inter1, FAB inter2); 
FAB cond_and_fab_rel(FAB inter1, FAB inter2, int eval_as_rel); 
FAB cond_or_fab(FAB inter1, FAB inter2);
FAB cond_or_fab_rel(FAB inter1, FAB inter2, int eval_as_rel);
FAB cond_if_fab(FAB inter1, FAB inter2);
FAB cond_iff_fab(FAB inter1, FAB inter2);

/* Returns Beran's number of the expression fval */
int MO2_to_Ber(FAB fval) ;

/* Constructs FAB element from Beran's number */
FAB Ber_to_MO2(int ber_val) ;
FAB Ber_to_MO2_ext(int ber_val, FAB a_aval, FAB b_val);

/* Symmetric differences: 
 *  Dorfer: Non-commutative Symmetric Differences in OML 
 *  
 */

/* (x ^ y') v ( x' ^ y) */
FAB  sdd_fab(FAB f1, FAB f2);
/* (x v y) ^ (x' v y') */
FAB  sdu_fab(FAB f1, FAB f2);
/* (x v (x' ^ y)) ^ (x' v y') */
FAB  sdl_fab(FAB f1, FAB f2);
/* ((x ^ y') v y) ^ (x' v y') */
FAB  sdr_fab(FAB f1, FAB f2);
/* (x v y) ^ (x' v (x ^ y')) */
FAB sdlp_fab(FAB f1, FAB f2);
/* (x v y) ^ ((x' ^ y) v y') */
FAB sdrp_fab(FAB f1, FAB f2);
#endif
