#ifndef _FABC_H_
#define _FABC_H_
#include "fab.h"

typedef struct {
	unsigned int f1:8, f2:8;
/*FAB f1, f2;*/
} FABC;

/* Constants */

extern const FABC Oc, Ac, cAc, Bc, cBc, Cc, cCc, Ic;
/* Boolean values */
extern const FABC Tc, Fc;

int is_bool_fabc(FABC f);

FABC addapt_bool_fabc(int i); 

/* Operations */
/* Nullary - constants */

/* Zero */
FABC O_fabc(void);
/* Element a - 1st element of the pair of non-commutative elements */
FABC A_fabc(void);
/* Element b - 2nd element of the pair of non-commutative elements */
FABC B_fabc(void);
/* Element c - element compatible with every element */
FABC C_fabc(void);
/* Unit */
FABC I_fabc(void);

/* complement */
FABC c_fabc(FABC f);

/* join */
FABC j_fabc(FABC f1, FABC f2);

/* meet */
FABC m_fabc(FABC f1, FABC f2);

/* skew join */
FABC sj_fabc(FABC f1, FABC f2);

/* skew meet */
FABC sm_fabc(FABC f1, FABC f2);


/* Sasaki projection */
FABC sp_fabc(FABC f1, FABC f2);

/* compatibility */
/* prints f1 and (f1/\f2) \/ (f1/\f2') */
void C_pr_fabc(FABC f1, FABC f2);
/* returns second expression (rs = right side) */
FABC C_rs_fabc(FABC f1, FABC f2);
/* return 1 if f1 C f2, otherwise returns 0 */
int C_i_fabc(FABC f1, FABC f2);
/* boolean FABC representation of the value C1_i */
FABC C_f_fabc(FABC f1, FABC f2);

/* Dual compatibility */
/* prints f1 and (f1\/f2) /\ (f1\/f2') */
void D_pr_fabc(FABC f1, FABC f2);
/* returns second expression (rs = right side)*/
FABC D_rs_fabc(FABC f1, FABC f2);
/* return 1 if f1 D f2, otherwise returns 0 */
int D_i_fabc(FABC f1, FABC f2);
/* boolean FABC representation of the value D1_i */
FABC D_f_fabc(FABC f1, FABC f2);

/*Implications*/
FABC imp0_fabc(FABC f1, FABC f2);
FABC eq0_fabc(FABC f1, FABC f2);
FABC imp1_fabc(FABC f1, FABC f2);
FABC eq1_fabc(FABC f1, FABC f2);
FABC imp2_fabc(FABC f1, FABC f2);
FABC eq2_fabc(FABC f1, FABC f2);
FABC imp3_fabc(FABC f1, FABC f2);
FABC eq3_fabc(FABC f1, FABC f2);
FABC imp4_fabc(FABC f1, FABC f2);
FABC eq4_fabc(FABC f1, FABC f2);
FABC imp5_fabc(FABC f1, FABC f2);
FABC eq5_fabc(FABC f1, FABC f2);

/* Upper commutator */
FABC uc_fabc(FABC f1, FABC f2);
	
/* Lower commutator */
FABC lc_fabc(FABC f1, FABC f2);

/* Inequality relations */

/* Equality */
int is_eq_fabc(FABC f1, FABC f2); 

FABC is_eq_f_fabc(FABC f1, FABC f2); 

int is_eq_inter_fabc(FABC f1, FABC f2, FABC interpretation);

FABC is_eq_inter_f_fabc(FABC f1, FABC f2, FABC interpretation);

/* less than or equal */
int is_leq_fabc(FABC f1, FABC f2); 

FABC is_leq_f_fabc(FABC f1, FABC f2); 

int is_leq_inter_fabc(FABC f1, FABC f2, FABC interpretation);

FABC is_leq_inter_f_fabc(FABC f1, FABC f2, FABC interpretation);

#if 0
/* less than */
int is_lt_fabc(FABC f1, FABC f2);

FABC is_lt_f_fabc(FABC f1, FABC f2); 

int is_lt_inter_fabc(FABC f1, FABC f2, FABC interpretation);

FABC is_lt_inter_f_fabc(FABC f1, FABC f2, FABC interpretation);
#endif

/* greater than or equal */
int is_geq_fabc(FABC f1, FABC f2); 

FABC is_geq_f_fabc(FABC f1, FABC f2); 

int is_geq_inter_fabc(FABC f1, FABC f2, FABC interpretation);

FABC is_geq_inter_f_fabc(FABC f1, FABC f2, FABC interpretation);

#if 0
/* greater than */
int is_gt_fabc(FABC f1, FABC f2);

FABC is_gt_f_fabc(FABC f1, FABC f2); 

int is_gt_inter_fabc(FABC f1, FABC f2, FABC interpretation);

FABC is_gt_inter_f_fabc(FABC f1, FABC f2, FABC interpretation);
#endif

/* Output functions*/

/* prints text-graphical representation of the element f */
void pr_fabc(FABC f);

/* prints TeX macro(foml2.sty) */
void tex_fabc(FABC f);

/* prints Beran's numbers of components */
void beran_fabc(FABC f);

/* prints both text-graphical and TeX macro representation */
void pt_fabc(FABC f);

/* prints text-graphical, TeX and Beran's number representation */
void ptb_fabc(FABC f);

/* Interpretations */

void create_interpretation_eq_fabc(FABC f1, FABC f2, FABC *interpretation); 
void create_interpretation_leq_fabc(FABC f1, FABC f2, FABC *interpretation); 
void create_interpretation_geq_fabc(FABC f1, FABC f2, FABC *interpretation); 
void create_interpretation_C_fabc(FABC f1, FABC f2, FABC *interpretation); 
void create_interpretation_D_fabc(FABC f1, FABC f2, FABC *interpretation); 

/* Conditional operations */

FABC cond_eq_fabc(FABC f1, FABC f2); 
FABC cond_eq_fabc_rel(FABC f1, FABC f2, int eval_as_rel); 
FABC cond_leq_fabc(FABC f1, FABC f2); 
FABC cond_leq_fabc_rel(FABC f1, FABC f2, int eval_as_rel); 
FABC cond_geq_fabc(FABC f1, FABC f2); 
FABC cond_geq_fabc_rel(FABC f1, FABC f2, int eval_as_rel); 
FABC cond_C_fabc(FABC f1, FABC f2); 
FABC cond_C_fabc_rel(FABC f1, FABC f2, int eval_as_rel); 
FABC cond_D_fabc(FABC f1, FABC f2); 
FABC cond_D_fabc_rel(FABC f1, FABC f2, int eval_as_rel); 
FABC cond_and_fabc(FABC inter1, FABC inter2); 
FABC cond_and_fabc_rel(FABC inter1, FABC inter2, int eval_as_rel); 
FABC cond_or_fabc(FABC inter1, FABC inter2);
FABC cond_or_fabc_rel(FABC inter1, FABC inter2, int eval_as_rel);
FABC cond_if_fabc(FABC inter1, FABC inter2);
FABC cond_iff_fabc(FABC inter1, FABC inter2);

FABC Ber_to_MO2_fabc(int ber_val); 
FABC Ber_to_MO2_fabc_ext(int ber_val, FABC arg_a, FABC arg_b); 

#endif

