#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dyn_string.h"

int init_string(Dynamic_str *ds) {
	return init_string_size(ds,CHUNK_SIZE - 1);	
}

int init_string_size(Dynamic_str *ds, int size) {
	int new_size;

	new_size = (size/CHUNK_SIZE+1)*CHUNK_SIZE;

	ds->str = (char *) malloc(new_size * sizeof(char));
	if (ds->str == NULL) return -1;

	ds->capacity=new_size;
	ds->str[0]='\0';
	ds->used=1;	

	return 0;
}

void release_string(Dynamic_str *ds) {
	if (ds->str) {
		free((void*)ds->str); 
		ds->str=NULL;
	}
	ds->used=0;
	ds->capacity=0;
}

int insert_string_at_position(Dynamic_str *ds, int position, char *astr) {
	int new_capacity, i, astrlen;
	char *pom;

	if (position>ds->used || position<0) {
		fprintf(stderr, "Error: Cannot insert string at position which is outside of the current string.\n");
		return -1;
	}
	if (position == ds->used)  /* Append at the end */
		return append_string(ds, astr);

	new_capacity = ds->used+strlen(astr);
	if (new_capacity>ds->capacity) {
		if (reallocate_dynamicstring(ds, new_capacity)==-1) return -1;	
	}

	astrlen = -strlen(astr); /* negative length of string astr */
	ds->str[ds->used-astrlen+1]='\0';

	i=0; 
	while (i>astrlen) {
		ds->str[ds->used-1-astrlen+i] = ds->str[ds->used-1+i];
		i--;
	}

	i=0; pom=astr;

	while (*pom) {
		ds->str[position+i]=*pom;
		i++;pom++;
	}
	ds->used -= astrlen; 
	return  0;
	
}
int reallocate_dynamicstring(Dynamic_str *ds, int desired_capacity) {
	int new_capacity=desired_capacity;
	
	new_capacity/=CHUNK_SIZE; new_capacity++;
	new_capacity*=CHUNK_SIZE;
	
	if (new_capacity != ds->capacity) {
		char *temp = ds->str, *pom;
		int i = 0;
		
		ds->str = (char *)malloc(sizeof(char) * new_capacity);

		if (ds->str == NULL) {
			fprintf(stderr,"Error: Not enough memory.\n");
			return -1;
		}

		pom = temp;
		while (*pom) {
			ds->str[i] = *pom;
			i++; pom++;
		}
		ds->str[i] = '\0';
		free((void *)temp);

		ds->capacity = new_capacity;
#ifdef DEBUG
	fprintf(stderr, "After reallocation: %s\n", ds->str);
#endif
	}
	return 0;
}

int append_string(Dynamic_str *ds, char *astr) {
#ifdef DEBUG

		fprintf(stderr, "\t\tAppending: `%s`\n", astr);
#endif
	int i;
	char *pom;

	if (ds->used + strlen(astr)>ds->capacity) {
#ifdef DEBUG

		fprintf(stderr, "\t\tNot enough capacity: `%s`\n", astr);
#endif
		if (reallocate_dynamicstring(ds, ds->used+strlen(astr))==-1) {
			return -1;
		}
	}
	i=0;pom=astr;
	while (*pom) {
		ds->str[ds->used-1+i]=*pom;
		i++; pom++;
	}
	if (i) ds->str[ds->used-1+i]='\0';
	ds->used+=strlen(astr);

	return 0;
}

/* TODO:*/
int remove_string(Dynamic_str *ds, int position, int size) {
	int pos,i;

	pos = position+size;
	if (pos>ds->used) pos = ds->used;
	i=position;

	while (i<=pos) {
		ds->str[i-size] = ds->str[i];
		i++;
	}
	return 0;
}
