#include <stdio.h>
#include "fab.h"
#include "beran_const_rpn.h"

extern FAB eval_RPN_ver3(char*, int);
extern FAB eval_RPN_in_ver3(char*, int,FAB,FAB);


/* Constants */
const FAB O = 0;	/* Zero element */
const FAB A = 92;	/* a  element   */
const FAB B = 58;	/* b  element   */
const FAB cA = 163;     /* a' element   */
const FAB cB = 197;     /* b' element   */
const FAB I  = 255;     /* One element  */
/* Boolean values */
const FAB T = 6;	/* True */
const FAB F = 9;	/* False */

/* TeX representation of the MO2 part(form2.sty) */
char *tex_repr[]={
/*       0     1     2      3    4     5     6     7 */
	"--", "--", "--", "-|", "--", "|-", "--", "||",
/*	 8     9    10     11    12   13     14   15 */
	"--", "--", "_|", "||", "|_", "||", "||", "||"
};

/* Normalize MO2 part */
int norm_tbl[] = {
0, 0, 0, 3, 0, 5, 0, 15, 0, 0, 10, 15, 12, 15, 15, 15
};

void normalize_fab(FAB *f) {
	*f = (*f & 240) | norm_tbl[*f & 15];
}

/* is boolean */
int is_bool_fab(FAB f) {
	return is_eq_fab(f, F) || is_eq_fab(f, T);
}

/* converts int bool value to T, F representation */
FAB addapt_bool_fab(int bool) {
	return (bool)?T:F;
}

/* Nullary operations - constants */
FAB O_fab(void) {return O;}
FAB A_fab(void) {return A;}
FAB B_fab(void) {return B;}
FAB I_fab(void) {return I;}

/* Join */
FAB j_fab(FAB f1, FAB f2) {
	FAB vysl;

	vysl = f1 | f2 ;
	normalize_fab(&vysl);
	return vysl;
}

/* Meet */
FAB m_fab(FAB f1, FAB f2) {
	FAB vysl;
	
	vysl = f1 & f2 ;
	normalize_fab(&vysl);
	return vysl;
}


/* Complement */
FAB c_fab(FAB f) {
	FAB vysl;
	
	vysl = 255-f;
	return vysl;
}

/***************************************************/
/* 	          Binary operations		   */
/***************************************************/

/*************/
/* Skew join */
/*************/
FAB sj_fab(FAB f1, FAB f2) {
	return 	j_fab(
			m_fab(
				f1,
				c_fab(f2)
			), 
			f2
		);
}

/*************/
/* Skew meet */
/*************/
FAB sm_fab(FAB f1, FAB f2) {
	return 	m_fab(
			j_fab( f1, c_fab(f2)), 
			f2
		);
}

/* Compatibility relation */
FAB C_rs_fab(FAB f1, FAB f2) {
	return 	j_fab(
			m_fab(f1,f2), 
			m_fab( f1, c_fab(f2))
		);
}

void C_pr_fab(FAB f1, FAB f2) {
	pr_fab(f1);
	pr_fab(C_rs_fab(f1,f2));
}


int C_i_fab(FAB f1, FAB f2) {
	return is_eq_fab(f1, C_rs_fab(f1, f2));
}

FAB C_f_fab(FAB f1, FAB f2) {
	return addapt_bool_fab(C_i_fab(f1,f2));
}

/* Dual compatibility relation */

FAB D_rs_fab(FAB f1, FAB f2) {
	return 	m_fab(
			j_fab(f1,f2), 
			j_fab( f1, c_fab(f2))
		);
}

void D_pr_fab(FAB f1, FAB f2) {
	pr_fab(f1);
	pr_fab(	D_rs_fab(f1,f2));
}

int D_i_fab(FAB f1, FAB f2) {
	return is_eq_fab(f1, D_rs_fab(f1, f2));
}

FAB D_f_fab(FAB f1, FAB f2) {
	return addapt_bool_fab(D_i_fab(f1, f2));
}

/* Sasaki projection */
FAB sp_fab(FAB f1, FAB f2) {
	return m_fab(
			f1, 
			j_fab(c_fab(f1), f2)
		);
}

/* Implications from_fab : 
	Quantumm_fab and Classical Implication Algebras with Primitive Implications
*/

/* classical */
FAB imp0_fab(FAB f1, FAB f2) {
	return j_fab(c_fab(f1), f2);
}

/* Sasaki */
FAB imp1_fab(FAB f1, FAB f2) {
	return j_fab(c_fab(f1), m_fab(f1,f2));
}

/* Dishkant */
FAB imp2_fab(FAB f1, FAB f2) {
	return j_fab(f2, m_fab(c_fab(f1), c_fab(f2)));
}

/* Kalmbach */
FAB imp3_fab(FAB f1, FAB f2) {
	return j_fab(
		j_fab(
		 m_fab(c_fab(f1), f2), 
		 m_fab(c_fab(f1), c_fab(f2))
                ), 
                m_fab(f1, j_fab(c_fab(f1), f2))
	       );
}

/* non-tollens */
FAB imp4_fab(FAB f1, FAB f2) {
	return j_fab(
		j_fab(
		 m_fab(f1,f2),
		 m_fab(c_fab(f1), f2)
		),
		m_fab(
		 j_fab(c_fab(f1),f2),
		 c_fab(f2)
		)
	       );
}

/* relevance */
FAB imp5_fab(FAB f1, FAB f2) {
	return j_fab(
		j_fab(
		 m_fab(f1,f2), 
		 m_fab(c_fab(f1),f2)
		),
		m_fab(c_fab(f1), c_fab(f2))
		);
}

/* Equivalences from Pavic, Megill */
FAB eq0_fab(FAB f1, FAB f2) {
	return m_fab(
		j_fab(c_fab(f1),f2), 
		j_fab(f1,c_fab(f2))
		);
}

FAB eq1_fab(FAB f1, FAB f2) {
	return m_fab(j_fab(f1,c_fab(f2)),j_fab(c_fab(f1),m_fab(f1,f2)));
}

FAB eq2_fab(FAB f1, FAB f2) {

	return m_fab(
		j_fab(f1,c_fab(f2)),
		j_fab(	f2,
			m_fab(c_fab(f1), c_fab(f2))
		)
	);
}
FAB eq3_fab(FAB f1, FAB f2) {
	return m_fab(j_fab(c_fab(f1),f2),j_fab(f1,m_fab(c_fab(f1),c_fab(f2))));
}
FAB eq4_fab(FAB f1, FAB f2) {
	return m_fab(j_fab(c_fab(f1),f2),j_fab(c_fab(f2),m_fab(f1,f2)));
}
FAB eq5_fab(FAB f1, FAB f2) {
	return j_fab(m_fab(f1,f2),m_fab(c_fab(f1),c_fab(f2)));
}

/* Lower commutator */
FAB lc_fab(FAB f1, FAB f2) {
	return 	j_fab(
			j_fab(
				j_fab(
					m_fab(f1,f2),
					m_fab(f1,c_fab(f2))
				 ),
				m_fab(c_fab(f1),f2)
			 ),
			m_fab(c_fab(f1),c_fab(f2))
		);
}

/* Upper commutator */
FAB uc_fab(FAB f1, FAB f2) {
	return 	m_fab(
			m_fab(
				m_fab(
					j_fab(f1,f2),
					j_fab(f1,c_fab(f2))
				 ),
				j_fab(c_fab(f1),f2)
			 ),
			j_fab(c_fab(f1),c_fab(f2))
		);
}

/* Inequality relations */

/* Equality */
int is_eq_fab(FAB f1, FAB f2) {
	return (f1 == f2);
}

FAB is_eq_f_fab(FAB f1, FAB f2) {
	return addapt_bool_fab(is_eq_fab(f1,f2));
}

int is_eq_inter_fab(FAB f1, FAB f2, FAB interpretation) {
	return is_eq_fab(
			m_fab(f1, interpretation),
			m_fab(f2, interpretation)
		);
}

FAB is_eq_inter_f_fab(FAB f1, FAB f2, FAB interpretation) {
	return addapt_bool_fab(is_eq_inter_fab(f1, f2, interpretation));
}

/* less than or equal */
int is_leq_fab(FAB f1, FAB f2) {
	return is_eq_fab(f1,m_fab(f1,f2));
}

FAB is_leq_f_fab(FAB f1, FAB f2) {
	return addapt_bool_fab(is_leq_fab(f1,f2));
}

int is_leq_inter_fab(FAB f1, FAB f2, FAB interpretation) {
	return is_leq_fab(
			m_fab(f1, interpretation),
			m_fab(f2, interpretation)
		);
}

FAB is_leq_inter_f_fab(FAB f1, FAB f2, FAB interpretation) {
	return addapt_bool_fab(is_leq_inter_fab(f1, f2, interpretation));
}


/* greater than or equal */
int is_geq_fab(FAB f1, FAB f2) {
	return is_eq_fab(f1,j_fab(f1,f2));
}

FAB is_geq_f_fab(FAB f1, FAB f2) {
	return addapt_bool_fab(is_geq_fab(f1,f2));
}

int is_geq_inter_fab(FAB f1, FAB f2, FAB interpretation) {
	return is_geq_fab(
			m_fab(f1, interpretation),
			m_fab(f2, interpretation)
		);
}

FAB is_geq_inter_f_fab(FAB f1, FAB f2, FAB interpretation) {
	return addapt_bool_fab(is_geq_inter_fab(f1, f2, interpretation));
}


#if 0
/* greater than */
int is_gt_fab(FAB f1, FAB f2) {
	return (is_geq_fab(f1, f2) && !is_eq_fab(f1, f2)) ;
}

FAB is_gt_f_fab(FAB f1, FAB f2) {
	return addapt_bool_fab(is_gt_fab(f1,f2));
}

int is_gt_inter_fab(FAB f1, FAB f2, FAB interpretation) {
	return is_gt_fab(
			m_fab(f1, interpretation),
			m_fab(f2, interpretation)
		);
}

FAB is_gt_inter_f_fab(FAB f1, FAB f2, FAB interpretation) {
	return addapt_bool_fab(is_gt_inter_fab(f1, f2, interpretation));
}
#endif

/* Output functions */

/*
      1
   -------
  |   o   |		128
4 | o   o | 2	     64	    32
  |   o   |		16
   -------
      8
*/

void prnthrow_fab(FAB f, int n) {
	switch(n) {
		case 1: prrow1st_fab(f); break;
		case 2: prrow2nd_fab(f); break;
		case 3: prrow3rd_fab(f); break;
		case 4: prrow4th_fab(f); break;
		case 5: prrow5th_fab(f); break;
		default: break;
	}
}

void prrow1st_fab(FAB f) {
	if (f & 1) printf (" ------- ");
	else       printf ("         ");
}

void prrow2nd_fab(FAB f) {
	if (f & 4) printf ("|"); else printf(" ");
	printf("   ");

	if (f & 128) printf("x"); else printf("o");
	printf("   ");

	if (f & 2) printf("|"); else printf(" "); 
}

void prrow3rd_fab(FAB f) {

	if (f & 4) printf ("|"); else printf(" ");
	printf(" ");

	if (f & 64) printf("x"); else printf("o");
	printf("   ");

	if (f & 32) printf("x"); else printf("o");
	printf(" ");

	if (f & 2) printf("|");  else printf(" ");

}

void prrow4th_fab(FAB f) {

	if (f & 4) printf ("|"); else printf(" ");
	printf("   ");

	if (f & 16) printf("x"); else printf("o");
	printf("   ");

	if (f & 2) printf("|"); else printf(" ");
}

void prrow5th_fab(FAB f) {
	if (f & 8) printf(" ------- ");
	else	   printf("         ");
}

void pr_fab(FAB f) {
	if (is_bool_fab(f)) {
		printf( is_eq_fab(f,F)?"False":"True"); printf("\n");
	}
	else {
		prrow1st_fab(f); printf("\n");
		prrow2nd_fab(f); printf("\n");
		prrow3rd_fab(f); printf("\n");
		prrow4th_fab(f); printf("\n");
		prrow5th_fab(f); printf("\n");
	}
}

void tex_fab(FAB f) {
	if (!is_bool_fab(f)) {
		printf("\\nc%d%d%d%d%s",
			(f & (1<<7)) >> 7,
			(f & (1<<6)) >> 6,
			(f & (1<<5)) >> 5,
			(f & (1<<4)) >> 4,
			tex_repr[f & 15]);
	}
}	

void beran_fab(FAB f) {
	if (!is_bool_fab(f)) {
		printf("Beran's number: %d\n", MO2_to_Ber(f));
	}
}

void pt_fab(FAB f) {
	pr_fab(f);
	tex_fab(f);printf("\n");
}

void ptb_fab(FAB f) {
	pt_fab(f);
	beran_fab(f);
}

/* Interpretations */

void create_interpretation_eq_fab(FAB f1, FAB f2, FAB *interpretation) {
	*interpretation=I;
	if ((f1 & 15) != (f2 & 15)) /* Ignore MO2 part */
		*interpretation &=240;
	if ( (f1 & (1<<7)) != (f2 & (1<<7)) ) *interpretation &= ~(1<<7);
	if ( (f1 & (1<<6)) != (f2 & (1<<6)) ) *interpretation &= ~(1<<6);
	if ( (f1 & (1<<5)) != (f2 & (1<<5)) ) *interpretation &= ~(1<<5);
	if ( (f1 & (1<<4)) != (f2 & (1<<4)) ) *interpretation &= ~(1<<4);
}

void create_interpretation_leq_fab(FAB f1, FAB f2, FAB *interpretation) {
	create_interpretation_eq_fab(f1, m_fab(f1, f2), interpretation);
}

void create_interpretation_geq_fab(FAB f1, FAB f2, FAB *interpretation) {
	create_interpretation_eq_fab(f1, j_fab(f1, f2), interpretation);
}

void create_interpretation_C_fab(FAB f1, FAB f2, FAB *interpretation) {
	create_interpretation_eq_fab(f1,C_rs_fab(f1,f2), interpretation);
}

void create_interpretation_D_fab(FAB f1, FAB f2, FAB *interpretation) {
	create_interpretation_eq_fab(f1,D_rs_fab(f1,f2), interpretation);
}

/* Conditional operations */
FAB cond_eq_fab(FAB f1, FAB f2) {
	FAB vysl;

	create_interpretation_eq_fab(f1,f2,&vysl);
	return vysl;
	
}

FAB cond_eq_fab_rel(FAB f1, FAB f2, int eval_as_rel) {
	FAB vysl;
	create_interpretation_eq_fab(f1,f2,&vysl);
	if (!eval_as_rel) 
		return vysl;
	
	return addapt_bool_fab(is_eq_fab(vysl,I));
}


FAB cond_leq_fab(FAB f1, FAB f2) {
	FAB vysl;
	
	create_interpretation_leq_fab(f1, f2, &vysl);
	return vysl;
}

FAB cond_leq_fab_rel(FAB f1, FAB f2, int eval_as_rel) {
	FAB vysl;
	
	create_interpretation_leq_fab(f1, f2, &vysl);
	if (!eval_as_rel)
		return vysl;

	return addapt_bool_fab(is_eq_fab(vysl,I));
}


FAB cond_geq_fab(FAB f1, FAB f2) {
	FAB vysl;
	
	create_interpretation_geq_fab(f1, f2, &vysl);
	return vysl;
}

FAB cond_geq_fab_rel(FAB f1, FAB f2, int eval_as_rel) {
	FAB vysl;
	
	create_interpretation_geq_fab(f1, f2, &vysl);
	if (!eval_as_rel)
		return vysl;

	return addapt_bool_fab(is_eq_fab(vysl,I));
}



FAB cond_C_fab(FAB f1, FAB f2) {
	FAB vysl;

	create_interpretation_C_fab(f1, f2, &vysl);
	return vysl;
}

FAB cond_C_fab_rel(FAB f1, FAB f2, int eval_as_rel) {
	FAB vysl;
	
	create_interpretation_C_fab(f1, f2, &vysl);
	if (!eval_as_rel)
		return vysl;
	else
		return addapt_bool_fab(is_eq_fab(vysl,I));
}



FAB cond_D_fab(FAB f1, FAB f2) {
	FAB vysl;

	create_interpretation_D_fab(f1, f2, &vysl);
	return vysl;
}

FAB cond_D_fab_rel(FAB f1, FAB f2, int eval_as_rel) {
	FAB vysl;
	
	create_interpretation_D_fab(f1, f2, &vysl);
	if (!eval_as_rel)
		return vysl;
	else
		return addapt_bool_fab(is_eq_fab(vysl,I));
}


FAB cond_and_fab(FAB inter1, FAB inter2) {
	return m_fab(inter1, inter2); /* MO2 part can be only 0000 or 1111 */
}

FAB cond_and_fab_rel(FAB inter1, FAB inter2, int eval_as_rel) {
	if (!eval_as_rel)
		return m_fab(inter1, inter2); /* MO2 part can be only 0000 or 1111 */
	return addapt_bool_fab(is_eq_fab(inter1,T) && is_eq_fab(inter2,T));
}

FAB cond_or_fab(FAB inter1, FAB inter2) {
	return j_fab(inter1, inter2);
}

FAB cond_or_fab_rel(FAB inter1, FAB inter2, int eval_as_rel) {
	if (!eval_as_rel)
		return j_fab(inter1, inter2);

	return addapt_bool_fab(is_eq_fab(inter1,T) || is_eq_fab(inter2,T));
}

FAB cond_if_fab(FAB inter1, FAB inter2) {
	return is_leq_f_fab(inter1, inter2);
}

FAB cond_iff_fab(FAB inter1, FAB inter2) {
	return is_eq_f_fab(inter1,inter2);
}

int Ber1_tr(int val) {
	switch (val) {
		case 0:
			return 0;
		case 3:
			return 4;
		case 5:
			return 3;
		case 10:
			return 2;
		case 12:
			return 1;
		case 15:
			return 5;
		default:
			fprintf(stderr, "Warning: Transformin unknown value (%d). Only 0,3,5,10,12,15 are allowed.\n", val);
			return -1;	
	}
}	

int inv_Ber1_tr(int val) {
	switch(val) {
		case 0:
			return 0;
		case 1:
			return 12;
		case 2:
			return 10;
		case 3:
			return 5;
		case 4:
			return 3;
		case 5:
			return 15;
		default:
			fprintf(stderr, "Warning: Transforming unknow value (%d). Only value from 0 to 5 are allowed.",val);
			return -1;
	}
}


int Ber2_tr(int val) {
	switch (val) {
		case 0: return 0;
		case 1: return 1;
		case 2: return 3;
		case 3: return 6;
		case 4: return 2;
		case 5: return 5;
		case 6: return 8;
		case 7: return 11;
		case 8: return 4;
		case 9: return 7;
		case 10: return 10;
		case 11: return 13;
		case 12: return 9;
		case 13: return 12;
		case 14: return 14;
		case 15: return 15;
		default:
			 fprintf(stderr, "Warning: Unknown value in transformation. Only value from 0 to 15 is expected.\n");
			 return -1;
	}
}

int inv_Ber2_tr(int val) {
	switch (val) {
		case 0: 
		case 1:
		case 5:
		case 10:
		case 14:
		case 15: return val;
		case 2: return 4;
		case 3: return 2;
		case 4: return 8;
		case 6: return 3;
		case 7: return 9;
		case 8: return 6;
		case 9: return 12;
		case 11: return 7;
		case 12: return 13;
		case 13: return 11;
		default:
			 fprintf(stderr, "Warning: Unknown value in transformation. Only value from 0 to 15 is expected.\n");
			 return -1;
	}
}

/* Returns Beran's value of the MO2 element */
int MO2_to_Ber(FAB fval) {
	return Ber2_tr((fval & 240)/16) + 16*Ber1_tr(fval & 15) + 1; 	
}

/* Returns the MO2 element from Beran's value */
FAB Ber_to_MO2_old(int ber_val) {
	if (ber_val<1 || ber_val>96) {
		fprintf(stderr,"Error: Invalid Beran's number (%d). Only value 1-96 are allowed.\n", ber_val);
		return O;
	}
	return 16*inv_Ber2_tr((ber_val-1) % 16) + inv_Ber1_tr((ber_val-1)/16);
}

FAB Ber_to_MO2(int ber_val) {
	return eval_RPN_ver3((char *)beran_const_rpn[ber_val-1],0);
}
FAB Ber_to_MO2_ext(int ber_val, FAB a_val, FAB b_val) {
	return eval_RPN_in_ver3((char *)beran_const_rpn[ber_val-1], a_val, b_val,0);
}

/* Symmetric differences - Dorfer: Non-Commutative Symmetric Differences */

FAB sdd_fab(FAB f1, FAB f2) {
	return j_fab(
			m_fab(
				f1,
				c_fab(f2)
			),
			m_fab(
				c_fab(f1),
				f2
			)
		);
}

FAB sdu_fab(FAB f1, FAB f2) {
	return 	m_fab(
			j_fab(
				f1, 
				f2
			), 
			j_fab(
				c_fab(f1), 
				c_fab(f2)
			)
		);
}

FAB sdl_fab(FAB f1, FAB f2) {
	return m_fab(
			j_fab(
				f1,
				m_fab(
					c_fab(f1), 
					f2
				)
			),
			j_fab(
				c_fab(f1),
				c_fab(f2)
			)
		);
}

FAB sdr_fab(FAB f1, FAB f2) {
	return m_fab(
			j_fab(
				m_fab(
					f1, 
					c_fab(f2)
				), 
				f2
			), 
			j_fab(
				c_fab(f1), 
				c_fab(f2)
			)
		);
}

FAB sdlp_fab(FAB f1, FAB f2) {
	return m_fab(
			j_fab(
				f1, 
				f2
			), 
			j_fab(
				c_fab(f1), 
				m_fab(
					f1, 
					c_fab(f2)
				)
			)
		);
}

FAB sdrp_fab(FAB f1, FAB f2) {
	return 	m_fab(
			j_fab(
				f1,
				f2
			),
			j_fab(
				m_fab(
					c_fab(f1), 
					f2
				),
				c_fab(f2)
			)
		);
}
