#include <stdio.h>
#include <stdlib.h>
#include "fab.h"
#include "fab_dynstack.h"

/* DYNSTACK*/
void init_dynstack_fab(DYNSTACK_FAB *st) {
	st->top=st->bottom=NULL;
}

int is_empty_dynstack_fab(DYNSTACK_FAB *st) {
	return st->bottom==st->top;
}

int push_dynstack_with_value_fab(DYNSTACK_FAB *st, FAB element, int value) {
	FAB_LIST *pom;

	pom = (FAB_LIST *)malloc(sizeof(FAB_LIST));
	if (pom == NULL) {
		fprintf(stderr, "Not enough memory: Could not allocate memory for struct FAB_LIST.\n");
		return -1;
	}	

	pom->next=st->top;
	pom->element=element;
	pom->semantic_value=value;
	st->top=pom;

	return 0;

}
int push_dynstack_fab(DYNSTACK_FAB *st, FAB element) {
	
	FAB_LIST *pom;

	pom = (FAB_LIST *)malloc(sizeof(FAB_LIST));
	if (pom == NULL) {
		fprintf(stderr, "Not enough memory: Could not allocate memory for struct FAB_LIST.\n");
		return -1;
	}	

	pom->next=st->top;
	pom->element=element;
	pom->semantic_value=0;
	st->top=pom;

	return 0;
}

int pop_dynstack_with_value_fab(DYNSTACK_FAB *st, FAB *element, int *value) {
	FAB_LIST *pom;

	if (is_empty_dynstack_fab(st)) {
		fprintf(stderr, "Stack is empty. (Error in the epression.)\n");
		return -1;
	}

	*element = st->top->element;
	*value=st->top->semantic_value;
	pom=st->top->next;
	free((void *)st->top);
	st->top=pom;

	return 0;

}

int pop_dynstack_fab(DYNSTACK_FAB *st, FAB *element) {
	FAB_LIST *pom;

	if (is_empty_dynstack_fab(st)) {
		fprintf(stderr, "Stack is empty. (Error in the epression.)\n");
		return -1;
	}

	*element = st->top->element;
	pom=st->top->next;
	free((void *)st->top);
	st->top=pom;

	return 0;
}

void flush_dynstack_fab(DYNSTACK_FAB *st) {
	FAB elem;

	while (!is_empty_dynstack_fab(st)) {
		pop_dynstack_fab(st, &elem);
	}
}

/* NEW_DYNSTACK */

void init_new_dynstack_fab(NEW_DYNSTACK_FAB *st) {
	st->top=st->bottom=NULL;
}

int is_empty_new_dynstack_fab(NEW_DYNSTACK_FAB *st) {
	return st->bottom==st->top;
}

int push_fab_new_dynstack_fab(NEW_DYNSTACK_FAB *st, FAB f_el) {
	t_stack_element el;
	el.f_element=f_el;
	return push_new_dynstack_fab(st, &el, FAB_TYPE);
}

int push_int_new_dynstack_fab(NEW_DYNSTACK_FAB *st, int i_el) {
	t_stack_element el;
	el.i_element=i_el;
	return push_new_dynstack_fab(st, &el, INT_TYPE);
}

int push_new_dynstack_fab(NEW_DYNSTACK_FAB *st, t_stack_element *el, t_element_type type) {
	NEW_FAB_LIST *pom;

	pom = (NEW_FAB_LIST *)malloc(sizeof(NEW_FAB_LIST));
	if (pom == NULL) {
		fprintf(stderr, "Not enough memory: Could not allocate memory for struct NEW_FAB_LIST.\n");
		return -1;
	}	

	switch (type) {
		case INT_TYPE:
			pom->el.i_element=el->i_element;
			break;
		case FAB_TYPE:
			pom->el.f_element=el->f_element;
			break;
		default:
			fprintf(stderr, "Error: Unknown stack type!\n");
			free((void *)pom);
			return -1;
			break;
	}
	pom->element_type=type;
	pom->next=st->top;
	st->top=pom;

	return 0;
}

int pop_int_new_dynstack_fab(NEW_DYNSTACK_FAB *st, int *value) {
	int status;
	t_stack_element el;
	t_element_type type;
	status = pop_new_dynstack_fab(st, &el, &type);
	if (status < 0) return -1;
	if (type != INT_TYPE) {
		fprintf(stderr, "Error: Not integer value in stack.\n");
		return -1;
	}
	*value = el.i_element;
	return 0;
}

int pop_fab_new_dynstack_fab(NEW_DYNSTACK_FAB *st, FAB *value) {
	int status;
	t_stack_element el;
	t_element_type type;

	status = pop_new_dynstack_fab(st, &el, &type);
	if (status < 0) return -1;
	if (type != FAB_TYPE) {
		fprintf(stderr, "Error: Not FAB value in stack.\n");
		return -1;
	}
	*value = el.f_element;
	return 0;
}

int pop_new_dynstack_fab(NEW_DYNSTACK_FAB *st, t_stack_element *el, t_element_type *type) {
	NEW_FAB_LIST *pom;

	if (is_empty_new_dynstack_fab(st)) {
		fprintf(stderr, "Stack is empty. (Error in the epression.)\n");
		return -1;
	}

	switch (st->top->element_type) {
		case INT_TYPE:
			el->i_element=st->top->el.i_element;
			break;
		case FAB_TYPE:
			el->f_element=st->top->el.f_element;
			break;
		default:
			fprintf(stderr, "Error: Unknown stack type.\n");
			return -1;
	}
	*type=st->top->element_type;

	pom=st->top->next;
	free((void *)st->top);
	st->top=pom;
	return 0;
}

void flush_new_dynstack_fab(NEW_DYNSTACK_FAB *st) {
	t_stack_element el;
	t_element_type type;

	while (!is_empty_new_dynstack_fab(st)) {
		pop_new_dynstack_fab(st, &el, &type);
	}
}

