#include <stdio.h>
#include <string.h>
#include "fab.h"
#include "fab_rpneval.h"
#include "fabx_parser.h"
#include "fabx_rpn.h"

R_ELEM_FAB operations[]={
	{OP_ERR,	   NONE_TYPE,	NULL},
	{OP_NONE,	   NONE_TYPE,	NULL},
	{ARG_O,		   TYPE,	(FAB (*)()) &O_fab},
	{ARG_A,		   TYPE,	(FAB (*)()) &A_fab},
	{ARG_B,		   TYPE,	(FAB (*)()) &B_fab},
	{ARG_I,		   TYPE,	(FAB (*)()) &I_fab},
	{OP_JOIN,	FF_TYPE,	(FAB (*)()) &j_fab},
	{OP_MEET,	FF_TYPE,	(FAB (*)()) &m_fab},
	{OP_COMPLEMENT,	 F_TYPE,	(FAB (*)()) &c_fab},
	{OP_SKEWJOIN,	FF_TYPE,	(FAB (*)()) &sj_fab},
	{OP_SKEWMEET,	FF_TYPE,	(FAB (*)()) &sm_fab},
	{OP_SASAKI,	FF_TYPE,	(FAB (*)()) &sp_fab},
	{OP_IMP0,	FF_TYPE,	(FAB (*)()) &imp0_fab},
	{OP_IMP1,	FF_TYPE,	(FAB (*)()) &imp1_fab},
	{OP_IMP2,	FF_TYPE,	(FAB (*)()) &imp2_fab},
	{OP_IMP3,	FF_TYPE,	(FAB (*)()) &imp3_fab},
	{OP_IMP4,	FF_TYPE,	(FAB (*)()) &imp4_fab},
	{OP_IMP5,	FF_TYPE,	(FAB (*)()) &imp5_fab},
	{OP_C,		FF_TYPE,	(FAB (*)()) &C_rs_fab},
	{OP_D,		FF_TYPE,	(FAB (*)()) &D_rs_fab},
	{COND_EQ__REL,	FF_I_TYPE,	(FAB (*)()) &cond_eq_fab_rel},
	{COND_LEQ_REL,	FF_I_TYPE,	(FAB (*)()) &cond_leq_fab_rel},
	{COND_GEQ_REL,	FF_I_TYPE,	(FAB (*)()) &cond_geq_fab_rel},
	{COND_RC__REL,	FF_I_TYPE,	(FAB (*)()) &cond_C_fab_rel},
	{COND_RD__REL,	FF_I_TYPE,	(FAB (*)()) &cond_D_fab_rel},
	{OP_IF_,	 FF_TYPE,	(FAB (*)()) &cond_if_fab},
	{OP_IFF,	 FF_TYPE,	(FAB (*)()) &cond_iff_fab},
	{OP_AND_REL,	FF_I_TYPE,	(FAB (*)()) &cond_and_fab_rel},
	{OP_OR__REL,	FF_I_TYPE,	(FAB (*)()) &cond_or_fab_rel},
	{OP_LCOM,	 FF_TYPE,	(FAB (*)()) &lc_fab},
	{OP_UCOM,	 FF_TYPE,	(FAB (*)()) &uc_fab},
	{BER1,		  I_TYPE,	(FAB (*)()) &Ber_to_MO2},
	{BER3,		IFF_TYPE,	(FAB (*)()) &Ber_to_MO2_ext},
	/* Sentinel */	
	{SENT,		NONE_TYPE,	NULL} 
};

#if 0
FAB_OP find_fun(OPTYPE el, int *arity) {
	int i=0;

	while (operations[i].elem!=SENT) {
		if (operations[i].elem == el) {
			*arity=operations[i].arity;
			return operations[i].fun;
		}
		i++;
	}
	return NULL;
}
#endif 
FAB_OP new_find_fun(OPTYPE el, t_fun_type *f_type) {
	int i=0;

	while (operations[i].elem!=(OPTYPE)SENT) {
		if (operations[i].elem == el) {
			*f_type=operations[i].f_type;
			return operations[i].fun;
		}
		i++;
	}
	return NULL;
}


void evaluate_nullary_fab(DYNSTACK_FAB *fab, FAB (*op)()) {
	push_dynstack_fab(fab, (*op)());
}

void evaluate_TYPE_fab(NEW_DYNSTACK_FAB *fab, FAB (*op)()) {
	push_fab_new_dynstack_fab(fab, (*op)());
}

void evaluate_unary_fab(DYNSTACK_FAB *fab, FAB (*op)(FAB)) {
	FAB vysl, arg1;

	pop_dynstack_fab(fab, &arg1);
	vysl = (*op)(arg1);
	push_dynstack_fab(fab, vysl);	
}

void evaluate_F_TYPE_fab(NEW_DYNSTACK_FAB *fab, FAB (*op)(FAB)) {
	FAB vysl, arg1;

	pop_fab_new_dynstack_fab(fab, &arg1);
	vysl = (*op)(arg1);
	push_fab_new_dynstack_fab(fab, vysl);	
}

void evaluate_I_TYPE_fab(NEW_DYNSTACK_FAB *fab, FAB (*op)(int)) {
	FAB vysl;
	int arg1;

	pop_int_new_dynstack_fab(fab, &arg1);
	vysl = (*op)(arg1);
	push_fab_new_dynstack_fab(fab, vysl);	
}


void evaluate_binary_fab(DYNSTACK_FAB *fab, FAB (*op)(FAB, FAB)) {
	FAB vysl,arg1, arg2;

	pop_dynstack_fab(fab,&arg2);
	pop_dynstack_fab(fab,&arg1);
	vysl=(*op)(arg1, arg2);
	push_dynstack_fab(fab, vysl);
}

void evaluate_FF_TYPE_fab(NEW_DYNSTACK_FAB *fab, FAB (*op)(FAB, FAB)) {
	FAB vysl,arg1, arg2;

	pop_fab_new_dynstack_fab(fab,&arg2);
	pop_fab_new_dynstack_fab(fab,&arg1);
	vysl=(*op)(arg1, arg2);
	push_fab_new_dynstack_fab(fab, vysl);
}

void evaluate_binary_with_int_fab(DYNSTACK_FAB *fab, FAB (*op)(FAB, FAB, int), int param) {
	FAB vysl,arg1, arg2;

	pop_dynstack_fab(fab,&arg2);
	pop_dynstack_fab(fab,&arg1);
	vysl=(*op)(arg1, arg2, param);
	push_dynstack_fab(fab, vysl);
}

void evaluate_FF_I_TYPE_fab(NEW_DYNSTACK_FAB *fab, FAB (*op)(FAB, FAB, int), int param) {
	FAB vysl,arg1, arg2;

	pop_fab_new_dynstack_fab(fab,&arg2);
	pop_fab_new_dynstack_fab(fab,&arg1);
	vysl=(*op)(arg1, arg2, param);
	push_fab_new_dynstack_fab(fab, vysl);
}

void evaluate_IFF_TYPE_fab(NEW_DYNSTACK_FAB *fab, FAB (*op)(int, FAB, FAB)) {
	FAB vysl, arg2, arg3;
	int arg1;

	pop_int_new_dynstack_fab(fab, &arg1);
	pop_fab_new_dynstack_fab(fab, &arg2);
	pop_fab_new_dynstack_fab(fab, &arg3);
	vysl = (*op)(arg1, arg2, arg3);
	push_fab_new_dynstack_fab(fab, vysl);
}

FAB eval_RPN_ver3(char *rpn, int eval_as_rel) {
	return eval_RPN_in_ver3(rpn, A,B, eval_as_rel);
}

FAB eval_RPN_in_ver3(char *rpn, FAB ai, FAB bi, int eval_as_rel) {
	char *pom=rpn, pom_buffer[MAX_TOKEN_SIZE];
	NEW_DYNSTACK_FAB st;
	FAB vysl;
	int stat,len/*, value*/;

	init_new_dynstack_fab(&st);

	if (*pom != '0') {
		fprintf(stderr, "Error: The number of generators doesn't match.\n");
		return F;
	}
	pom++;

	while (*pom) {

		stat=-2;len=0;
		while (*pom && stat==-2) {
			pom_buffer[len++]=*pom;pom_buffer[len]='\0';
			stat=find_rpn(pom_buffer);
			pom++;
		}
		switch (stat) {
			case ARG_INT:
				{
					int value=0;
					while (*pom>='0' && *pom<='9') {value=10*value+*pom-'0'; pom++;}
					pom++; /* skip_space*/
					push_int_new_dynstack_fab(&st, value);
				}
				break;
			case ARG_A:
				push_fab_new_dynstack_fab(&st, ai);
				break;
			case ARG_B:
				push_fab_new_dynstack_fab(&st, bi);
				break;
			default: {
		
				t_fun_type f_type = NONE_TYPE;
				FAB_OP fun;

				fun = new_find_fun(stat, &f_type);
				switch (f_type) {
					case TYPE:
						evaluate_TYPE_fab(&st, fun);
						break;
					case F_TYPE:
						evaluate_F_TYPE_fab(&st, (FAB (*)(FAB)) fun);
						break;
					case I_TYPE:
						evaluate_I_TYPE_fab(&st, (FAB (*)(int)) fun);
						break;
					case FF_TYPE:
						evaluate_FF_TYPE_fab(&st, (FAB (*)(FAB,FAB)) fun);
						break;
					case FF_I_TYPE:
						evaluate_FF_I_TYPE_fab(&st, (FAB (*)(FAB,FAB, int)) fun, eval_as_rel);
						break;
					case IFF_TYPE:
						evaluate_IFF_TYPE_fab(&st, (FAB (*)(int, FAB, FAB)) fun);
						break;
					default:
						break;
				}
			}

		}/*end switch */
		
	}
	pop_fab_new_dynstack_fab(&st, &vysl);
	if (!is_empty_new_dynstack_fab(&st)) {
		fprintf(stderr, "Logical error in expression. (Not all operands were used.)\n");
		flush_new_dynstack_fab(&st);
		return F;
	}
	flush_new_dynstack_fab(&st);
	return vysl;
}
