#include <stdio.h>
#include "fab.h"
#include "fabc.h"
#include "dyn_string.h"
#include "beran_const_rpn.h"

/* Definition of constants */
const FABC Oc = {0,0};
const FABC Ac = {92, 92};
const FABC cAc = {163,163};
const FABC Bc = {58, 58};
const FABC cBc = {197,197};
const FABC Cc = { 0, 255};
const FABC cCc = {255, 0};
const FABC Ic = {255, 255};

/* boolean */
const FABC Tc = {6,6};
const FABC Fc = {9,9};

int is_bool_fabc(FABC f) {
	return is_eq_fabc(f, Tc) || is_eq_fabc(f,Fc);
}

FABC addapt_bool_fabc(int i) {
	return (i)?Tc:Fc;
}

/* Operations */

/* Nullary - constants */

FABC O_fabc(void) {return Oc;}
FABC A_fabc(void) {return Ac;}
FABC B_fabc(void) {return Bc;}
FABC C_fabc(void) {return Cc;}
FABC I_fabc(void) {return Ic;}
/* Complement */
FABC c_fabc(FABC f) {
	FABC vysl;

	vysl.f1 = c_fab(f.f1);
	vysl.f2 = c_fab(f.f2);
	return vysl;
}

/* Join */
FABC j_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = j_fab(f1.f1, f2.f1);
	vysl.f2 = j_fab(f1.f2, f2.f2);
	return vysl;
}

/* Meet */
FABC m_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = m_fab(f1.f1, f2.f1);
	vysl.f2 = m_fab(f1.f2, f2.f2);
	return vysl;
}

/* Skew join */
FABC sj_fabc(FABC f1, FABC f2) {
	return 	j_fabc(
			m_fabc(f1, c_fabc(f2)), 
			f2
		);
}

/* Skew meet */
FABC sm_fabc(FABC f1, FABC f2) {
	return 	m_fabc(
			j_fabc(f1, c_fabc(f2)), 
			f2
		);
}

/* Sasaki projection */
FABC sp_fabc(FABC f1, FABC f2) {
	return m_fabc(
		f1,
		j_fabc(c_fabc(f1), f2)
	);
}

/* Compatibility relation */
FABC C_rs_fabc(FABC f1, FABC f2) {
	return 
		j_fabc(
			m_fabc(f1, f2),
			m_fabc(f1,c_fabc(f2))
		)
	;	
}

int C_i_fabc(FABC f1, FABC f2) {
	return is_eq_fabc(f1, C_rs_fabc(f1, f2));
}

FABC C_f_fabc(FABC f1, FABC f2) {
	return addapt_bool_fabc(C_i_fabc(f1,f2));
}

void C_pr_fabc(FABC f1, FABC f2) {
	pr_fabc(f1);
	pr_fabc(C_rs_fabc(f1,f2));
}

/* Dual compatibility relation */
FABC D_rs_fabc(FABC f1, FABC f2) {
	return 
		m_fabc(
			j_fabc(f1,f2),
			j_fabc(f1,c_fabc(f2))
		)
	;
}

int D_i_fabc(FABC f1, FABC f2) {
	return is_eq_fabc(f1, D_rs_fabc(f1, f2));
}

FABC D_f_fabc(FABC f1, FABC f2) {
	return addapt_bool_fabc(D_i_fabc(f1,f2));
}

void D_pr_fabc(FABC f1, FABC f2) {
	pr_fabc(f1);
	pr_fabc(	D_rs_fabc(f1,f2));
}


FABC imp0_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = imp0_fab(f1.f1, f2.f1);
	vysl.f2 = imp0_fab(f1.f2, f2.f2);
	return vysl;
}

FABC eq0_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = eq0_fab(f1.f1, f2.f1);
	vysl.f2 = eq0_fab(f1.f2, f2.f2);
	return vysl;
}

FABC imp1_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = imp1_fab(f1.f1, f2.f1);
	vysl.f2 = imp1_fab(f1.f2, f2.f2);
	return vysl;
}

FABC eq1_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = eq1_fab(f1.f1, f2.f1);
	vysl.f2 = eq1_fab(f1.f2, f2.f2);
	return vysl;
}

FABC imp2_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = imp2_fab(f1.f1, f2.f1);
	vysl.f2 = imp2_fab(f1.f2, f2.f2);
	return vysl;
}

FABC eq2_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = eq2_fab(f1.f1, f2.f1);
	vysl.f2 = eq2_fab(f1.f2, f2.f2);
	return vysl;
}

FABC imp3_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = imp3_fab(f1.f1, f2.f1);
	vysl.f2 = imp3_fab(f1.f2, f2.f2);
	return vysl;
}

FABC eq3_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = eq3_fab(f1.f1, f2.f1);
	vysl.f2 = eq3_fab(f1.f2, f2.f2);
	return vysl;
}

FABC imp4_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = imp4_fab(f1.f1, f2.f1);
	vysl.f2 = imp4_fab(f1.f2, f2.f2);
	return vysl;
}

FABC eq4_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = eq4_fab(f1.f1, f2.f1);
	vysl.f2 = eq4_fab(f1.f2, f2.f2);
	return vysl;
}

FABC imp5_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = imp5_fab(f1.f1, f2.f1);
	vysl.f2 = imp5_fab(f1.f2, f2.f2);
	return vysl;
}

FABC eq5_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1 = eq5_fab(f1.f1, f2.f1);
	vysl.f2 = eq5_fab(f1.f2, f2.f2);
	return vysl;
}

/* Upper commutator */
FABC uc_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1=uc_fab(f1.f1,f2.f1);
	vysl.f2=uc_fab(f1.f2,f2.f2);
	return vysl;
}

/* Lower commutator */
FABC lc_fabc(FABC f1, FABC f2) {
	FABC vysl;

	vysl.f1=lc_fab(f1.f1,f2.f1);
	vysl.f2=lc_fab(f1.f2,f2.f2);
	return vysl;
}

/* Inequality relations */

/* Equality */
int is_eq_fabc(FABC f1, FABC f2) {
	return is_eq_fab(f1.f1, f2.f1) && is_eq_fab(f1.f2, f2.f2);
}

FABC is_eq_f_fabc(FABC f1, FABC f2) {
	return addapt_bool_fabc(is_eq_fabc(f1,f2));
}


int is_eq_inter_fabc(FABC f1, FABC f2, FABC interpretation) {
	return is_eq_fabc(
			m_fabc(f1, interpretation),
			m_fabc(f2, interpretation)
		);
}

FABC is_eq_inter_f_fabc(FABC f1, FABC f2, FABC interpretation) {
	return addapt_bool_fabc(is_eq_inter_fabc(f1, f2, interpretation));
}

/* less than or equal */
int is_leq_fabc(FABC f1, FABC f2) {
	return is_eq_fabc(f1,m_fabc(f1,f2));
}

FABC is_leq_f_fabc(FABC f1, FABC f2) {
	return addapt_bool_fabc(is_leq_fabc(f1,f2));
}

int is_leq_inter_fabc(FABC f1, FABC f2, FABC interpretation) {
	return is_leq_fabc(
			m_fabc(f1, interpretation),
			m_fabc(f2, interpretation)
		);
}

FABC is_leq_inter_f_fabc(FABC f1, FABC f2, FABC interpretation) {
	return addapt_bool_fabc(is_leq_inter_fabc(f1, f2, interpretation));
}


#if 0
/* less than */
int is_lt_fabc(FABC f1, FABC f2) {
	return (is_leq_fabc(f1, f2) && !is_eq_fabc(f1, f2)) ;
}

FABC is_lt_f_fabc(FABC f1, FABC f2) {
	return addapt_bool_fabc(is_lt_fabc(f1,f2));
}

int is_lt_inter_fabc(FABC f1, FABC f2, FABC interpretation) {
	return is_lt_fabc(
			m_fabc(f1, interpretation),
			m_fabc(f2, interpretation)
		);
}

FABC is_lt_inter_f_fabc(FABC f1, FABC f2, FABC interpretation) {
	return addapt_bool_fabc(is_lt_inter_fabc(f1, f2, interpretation));
}
#endif

/* greater than or equal */
int is_geq_fabc(FABC f1, FABC f2) {
	return is_eq_fabc(f1,j_fabc(f1,f2));
}
FABC is_geq_f_fabc(FABC f1, FABC f2) {
	return addapt_bool_fabc(is_geq_fabc(f1,f2));
}

int is_geq_inter_fabc(FABC f1, FABC f2, FABC interpretation) {
	return is_geq_fabc(
			m_fabc(f1, interpretation),
			m_fabc(f2, interpretation)
		);
}

FABC is_geq_inter_f_fabc(FABC f1, FABC f2, FABC interpretation) {
	return addapt_bool_fabc(is_geq_inter_fabc(f1, f2, interpretation));
}


#if 0
/* greater than */
int is_gt_fabc(FABC f1, FABC f2) {
	return (is_geq_fabc(f1, f2) && !is_eq_fabc(f1, f2)) ;
}
FABC is_gt_f_fabc(FABC f1, FABC f2) {
	return addapt_bool_fabc(is_gt_fabc(f1,f2));
}

int is_gt_inter_fabc(FABC f1, FABC f2, FABC interpretation) {
	return is_gt_fabc(
			m_fabc(f1, interpretation),
			m_fabc(f2, interpretation)
		);
}

FABC is_gt_inter_f_fabc(FABC f1, FABC f2, FABC interpretation) {
	return addapt_bool_fabc(is_gt_inter_fabc(f1, f2, interpretation));
}
#endif

/* Output functions */

void pr_fabc(FABC f) {
	if (is_bool_fabc(f)) {
		printf(is_eq_fabc(f,Tc)?"True": "False"); 
		printf("\n");
	}
	else {
		prrow1st_fab(f.f1); printf("   ");prrow1st_fab(f.f2);printf("\n");
		prrow2nd_fab(f.f1); printf("   ");prrow2nd_fab(f.f2);printf("\n");
		prrow3rd_fab(f.f1); printf(" , ");prrow3rd_fab(f.f2);printf("\n");
		prrow4th_fab(f.f1); printf("   ");prrow4th_fab(f.f2);printf("\n");
		prrow5th_fab(f.f1); printf("   ");prrow5th_fab(f.f2);printf("\n");
	}
}

void tex_fabc(FABC f) {

	if (!is_bool_fabc(f)) {
		tex_fab(f.f1);
		printf(", ");
		tex_fab(f.f2);
	}
}

void pt_fabc(FABC f) {
	pr_fabc(f);
	tex_fabc(f);printf("\n");
}

void ptb_fabc(FABC f) {
	pr_fabc(f);
	tex_fabc(f);printf("\n");
	beran_fabc(f);
}

void beran_fabc(FABC f) {
	if (!is_bool_fabc(f)) {
		printf("Beran's numbers of components: %3d, %3d\n", MO2_to_Ber(f.f1), MO2_to_Ber(f.f2));
	}
}


/* Interpretations */

void create_interpretation_eq_fabc(FABC f1, FABC f2, FABC *interpretation) {
	FAB inter;

	create_interpretation_eq_fab(f1.f1, f2.f1, &inter);
	interpretation->f1=inter;
	create_interpretation_eq_fab(f1.f2, f2.f2, &inter);
	interpretation->f2=inter;
}

void create_interpretation_leq_fabc(FABC f1, FABC f2, FABC *interpretation) {
	create_interpretation_eq_fabc(f1, m_fabc(f1, f2), interpretation);
}

void create_interpretation_geq_fabc(FABC f1, FABC f2, FABC *interpretation) {
	create_interpretation_eq_fabc(f1, j_fabc(f1, f2), interpretation);
}

void create_interpretation_C_fabc(FABC f1, FABC f2, FABC *interpretation) {
	create_interpretation_eq_fabc(f1,C_rs_fabc(f1,f2), interpretation);
}

void create_interpretation_D_fabc(FABC f1, FABC f2, FABC *interpretation) {
	create_interpretation_eq_fabc(f1,D_rs_fabc(f1,f2), interpretation);
}

/* Conditional operations */
FABC cond_eq_fabc(FABC f1, FABC f2) {
	FABC vysl;

	create_interpretation_eq_fabc(f1,f2,&vysl);
	return vysl;
}

FABC cond_eq_fabc_rel(FABC f1, FABC f2, int eval_as_rel) {
	FABC vysl;

	create_interpretation_eq_fabc(f1,f2,&vysl);
	if (!eval_as_rel)
		return vysl;

	return addapt_bool_fabc(is_eq_fabc(vysl, Ic));
}

FABC cond_leq_fabc(FABC f1, FABC f2) {
	FABC vysl;

	create_interpretation_leq_fabc(f1,f2,&vysl);
	return vysl;
}

FABC cond_leq_fabc_rel(FABC f1, FABC f2, int eval_as_rel) {
	FABC vysl;

	create_interpretation_leq_fabc(f1,f2,&vysl);
	if (!eval_as_rel)
		return vysl;
	return addapt_bool_fabc(is_eq_fabc(vysl, Ic));
}
FABC cond_geq_fabc(FABC f1, FABC f2) {
	FABC vysl;

	create_interpretation_geq_fabc(f1,f2,&vysl);
	return vysl;
}

FABC cond_geq_fabc_rel(FABC f1, FABC f2, int eval_as_rel) {
	FABC vysl;

	create_interpretation_geq_fabc(f1,f2,&vysl);
	if (!eval_as_rel)
		return vysl;
	return addapt_bool_fabc(is_eq_fabc(vysl, Ic));
}

FABC cond_C_fabc(FABC f1, FABC f2) {
	FABC vysl;

	create_interpretation_C_fabc(f1,f2,&vysl);
	return vysl;
}

FABC cond_C_fabc_rel(FABC f1, FABC f2, int eval_as_rel) {
	FABC vysl;

	create_interpretation_C_fabc(f1,f2,&vysl);
	if (!eval_as_rel)
		return vysl;
	return addapt_bool_fabc(is_eq_fabc(vysl, Ic));
}




FABC cond_D_fabc(FABC f1, FABC f2) {
	FABC vysl;

	create_interpretation_D_fabc(f1,f2,&vysl);
	return vysl;
}

FABC cond_D_fabc_rel(FABC f1, FABC f2, int eval_as_rel) {
	FABC vysl;

	create_interpretation_D_fabc(f1,f2,&vysl);
	if (!eval_as_rel)
		return vysl;
	return addapt_bool_fabc(is_eq_fabc(vysl, Ic));
}

FABC cond_and_fabc(FABC inter1, FABC inter2) {
/* In interpretations: MO2 part can be only 0000 or 1111 */
	return m_fabc(inter1, inter2); 
}

FABC cond_and_fabc_rel(FABC inter1, FABC inter2, int eval_as_rel) {
/* In interpretations: MO2 part can be only 0000 or 1111 */
	if (!eval_as_rel)
		return m_fabc(inter1, inter2); 
	return addapt_bool_fabc(is_eq_fabc(inter1, Tc) && is_eq_fabc(inter2, Tc));
}

FABC cond_or_fabc(FABC inter1, FABC inter2) {
/* In interpretations: MO2 part can be only 0000 or 1111 */
	return j_fabc(inter1, inter2); 
}

FABC cond_or_fabc_rel(FABC inter1, FABC inter2, int eval_as_rel) {
/* In interpretations: MO2 part can be only 0000 or 1111 */
	if (!eval_as_rel)
		return m_fabc(inter1, inter2); 
	return addapt_bool_fabc(is_eq_fabc(inter1, Tc) || is_eq_fabc(inter2, Tc));
}


#if 0
FABC cond_not_fabc(FABC inter) {
	return c_fabc(inter);
}
#endif

FABC cond_if_fabc(FABC inter1, FABC inter2) {
	return is_leq_f_fabc(inter1, inter2);
}

FABC cond_iff_fabc(FABC inter1, FABC inter2) {
	return is_eq_f_fabc(inter1,inter2);
}

#if 0
FABC cond_lt_fabc(FABC f1, FABC f2) {
	return cond_and_fabc(
			cond_leq_fabc(f1,f2), 
			cond_not_fabc(
				cond_eq_fabc(f1,f2)
			)
		);
}

FABC cond_gt_fabc(FABC f1, FABC f2) {
	return cond_and_fabc(
			cond_geq_fabc(f1,f2), 
			cond_not_fabc(
				cond_eq_fabc(f1,f2)
			)
		);
}
#endif
extern FABC eval_RPN_fabc_in_ver3(char *rpn, FABC ai, FABC bi, FABC ci, int eval_as_rel); 
FABC Ber_to_MO2_fabc(int ber_val) {
	return Ber_to_MO2_fabc_ext(ber_val, Ac, Bc);
}
FABC Ber_to_MO2_fabc_ext(int ber_val, FABC arg_a, FABC arg_b) {
	Dynamic_str rpn_copy;
	FABC vysl;

	init_string(&rpn_copy);
	append_string(&rpn_copy, (char *)beran_const_rpn[ber_val-1]);
	rpn_copy.str[0]='1';
	vysl = eval_RPN_fabc_in_ver3(rpn_copy.str, arg_a, arg_b, Cc, 0);
	release_string(&rpn_copy);
	return vysl;
}

