#include <stdio.h>
#include <stdlib.h>
#include "fabc.h"
#include "fabc_dynstack.h"


void init_dynstack_fabc(DYNSTACK_FABC *st) {
	st->top=st->bottom=NULL;
}

int is_empty_dynstack_fabc(DYNSTACK_FABC *st) {
	return st->bottom==st->top;
}

int push_dynstack_fabc(DYNSTACK_FABC *st, FABC element) {
	
	FABC_LIST *pom;

	pom = (FABC_LIST *)malloc(sizeof(FABC_LIST));
	if (pom == NULL) {
		fprintf(stderr, "Not enough memory: Could not allocate memory for struct FABC_LIST.\n");
		return -1;
	}	

	pom->next=st->top;
	pom->element=element;
	st->top=pom;

	return 0;
}

int pop_dynstack_fabc(DYNSTACK_FABC *st, FABC *element) {
	FABC_LIST *pom;

	if (is_empty_dynstack_fabc(st)) {
		fprintf(stderr, "Stack is empty. (Error in the epression.)\n");
		return -1;
	}

/*	pom=st->top;*/
	*element = st->top->element;
	pom=st->top->next;
	free((void *)st->top);
	st->top=pom;

	return 0;
}

void flush_dynstack_fabc(DYNSTACK_FABC *st) {
	FABC elem;

	while (!is_empty_dynstack_fabc(st)) {
		pop_dynstack_fabc(st, &elem);
	}
}

/* NEW_DYNSTACK */

void init_new_dynstack_fabc(NEW_DYNSTACK_FABC *st) {
	st->top=st->bottom=NULL;
}

int is_empty_new_dynstack_fabc(NEW_DYNSTACK_FABC *st) {
	return st->bottom==st->top;
}

int push_fabc_new_dynstack_fabc(NEW_DYNSTACK_FABC *st, FABC fc_el) {
	t_stack_element_fabc el;
	el.fc_element=fc_el;
	return push_new_dynstack_fabc(st, &el, FABC_TYPE);
}

int push_int_new_dynstack_fabc(NEW_DYNSTACK_FABC *st, int i_el) {
	t_stack_element_fabc el;
	el.i_element=i_el;
	return push_new_dynstack_fabc(st, &el, INT_TYPE);
}

int push_new_dynstack_fabc(NEW_DYNSTACK_FABC *st, t_stack_element_fabc *el, t_element_type_fabc type) {
	NEW_FABC_LIST *pom;

	pom = (NEW_FABC_LIST *)malloc(sizeof(NEW_FABC_LIST));
	if (pom == NULL) {
		fprintf(stderr, "Not enough memory: Could not allocate memory for struct NEW_FABC_LIST.\n");
		return -1;
	}	

	switch (type) {
		case INT_TYPE:
			pom->el.i_element=el->i_element;
			break;
		case FABC_TYPE:
			pom->el.fc_element=el->fc_element;
			break;
		default:
			fprintf(stderr, "Error: Unknown stack type!\n");
			free((void *)pom);
			return -1;
			break;
	}
	pom->element_type=type;
	pom->next=st->top;
	st->top=pom;

	return 0;
}

int pop_int_new_dynstack_fabc(NEW_DYNSTACK_FABC *st, int *value) {
	int status;
	t_stack_element_fabc el;
	t_element_type_fabc type;
	status = pop_new_dynstack_fabc(st, &el, &type);
	if (status < 0) return -1;
	if (type != INT_TYPE) {
		fprintf(stderr, "Error: Not integer value in stack.\n");
		return -1;
	}
	*value = el.i_element;
	return 0;
}

int pop_fabc_new_dynstack_fabc(NEW_DYNSTACK_FABC *st, FABC *value) {
	int status;
	t_stack_element_fabc el;
	t_element_type_fabc type;

	status = pop_new_dynstack_fabc(st, &el, &type);
	if (status < 0) return -1;
	if (type != FABC_TYPE) {
		fprintf(stderr, "Error: Not FABC value in stack.\n");
		return -1;
	}
	*value = el.fc_element;
	return 0;
}

int pop_new_dynstack_fabc(NEW_DYNSTACK_FABC *st, t_stack_element_fabc *el, t_element_type_fabc *type) {
	NEW_FABC_LIST *pom;

	if (is_empty_new_dynstack_fabc(st)) {
		fprintf(stderr, "Stack is empty. (Error in the epression.)\n");
		return -1;
	}

	switch (st->top->element_type) {
		case INT_TYPE:
			el->i_element=st->top->el.i_element;
			break;
		case FABC_TYPE:
			el->fc_element=st->top->el.fc_element;
			break;
		default:
			fprintf(stderr, "Error: Unknown stack type.\n");
			return -1;
	}
	*type=st->top->element_type;

	pom=st->top->next;
	free((void *)st->top);
	st->top=pom;
	return 0;
}

void flush_new_dynstack_fabc(NEW_DYNSTACK_FABC *st) {
	t_stack_element_fabc el;
	t_element_type_fabc type;

	while (!is_empty_new_dynstack_fabc(st)) {
		pop_new_dynstack_fabc(st, &el, &type);
	}
}


