#include <stdio.h>
#include <stdlib.h>

#include "fab.h"
#include "fabn.h"
#include "beran_const_rpn.h"
#include "dyn_string.h"

/** Helper function 'exp2' **
 * 
 * Computes 2^n
 */
static const int exp2_table[]={1,2,4,8,16,32,64,128,256,512};

int exp2_old(int n) {
	int exp=1,i;

	for (i=0; i<n;i++) exp*=2;
	return exp;
}

int exp2(int n) {

	if (n<0 || n>9) {/* unsupported value */
		fprintf(stderr, "Warning: unsupported value. Only 0-9 are accepted.\n");
		return 0; 
	}
	return exp2_table[n];
}

int create_fabn(FABN *f, int n) {

	f->fabn= (FAB *)calloc(sizeof(FAB), exp2(n));
	if (f->fabn == NULL) {
		fprintf(stderr,"Memory allocation error. (create_fabn).\n");
		return -1;
	}
	f->n=n;
	return 0;
}

void clear_fabn(FABN *f) {
	if (f->fabn) {
		free((void *)f->fabn);
		f->fabn=NULL;
	}
	f->n=0;
}

int is_bool_fabn(FABN *f) {
	FABN temp;
	int result=0;

	create_fabn(&temp, f->n);
	
	T_fabn(&temp);
	if (is_eq_fabn(&temp,f))  result=1;
	else {
		F_fabn(&temp);
		if (is_eq_fabn(&temp,f)) result=1;
	}
	clear_fabn(&temp);
	return result;
}

void addapt_bool_fabn(int bool, FABN *f_acc) {
	if (bool) T_fabn(f_acc);
	else F_fabn(f_acc);
}


void T_fabn(FABN *f) {
	int i;

	for (i=0;i<exp2(f->n);i++)
		f->fabn[i]=T;
}

void F_fabn(FABN *f) {
	int i;

	for (i=0;i<exp2(f->n);i++)
		f->fabn[i]=F;
}

void O_fabn(FABN *f) {
	int i;	

	for (i=0;i<exp2(f->n);i++)
		f->fabn[i]=O;
}

void A_fabn(FABN *f) {
	int i;	

	for (i=0;i<exp2(f->n);i++)
		f->fabn[i]=A;
}

void B_fabn(FABN *f) {
	int i;	
	
	for (i=0;i<exp2(f->n);i++) 
		f->fabn[i]=B;
}

void Ci_fabn(FABN *f, int ith) {
	int mask = 1<<(f->n-ith), i;

	for (i=0;i<exp2(f->n);i++)
		f->fabn[i] = (i & mask)?I:O;
}

void I_fabn(FABN *f) {
	int i;	

	for (i=0;i<exp2(f->n);i++)
		f->fabn[i]=I;
}

void forall_bin_fabn(FABN *f_acc, FABN *f_op, FAB (*fun)(FAB,FAB)) {
	int i;

	for (i=0; i<exp2(f_acc->n);i++) 
		f_acc->fabn[i]=(*fun)(f_acc->fabn[i], f_op->fabn[i]);
}

void load_fabn(FABN *f_acc, FABN *f_op) {
	int i;
	
	/* TODO: Dangerous: if not enough memory was allocated. */
	f_acc->n = f_op->n;
	for (i=0; i<exp2(f_acc->n);i++) {
		f_acc->fabn[i]=f_op->fabn[i];
	}
}

void c_fabn(FABN *f_acc) {
	int i;

	for (i=0; i<exp2(f_acc->n);i++) 
		f_acc->fabn[i]=c_fab(f_acc->fabn[i]);
}


/* computes f_acc () f_op and stores it to f_acc */
void j_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &j_fab);
}

void m_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &m_fab);
}

void sj_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &sj_fab);
}

void sm_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &sm_fab);
}

void sp_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &sp_fab);
}

void C_rs_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &C_rs_fab);
}

void D_rs_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &D_rs_fab);
}

void imp0_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &imp0_fab);
}

void eq0_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &eq0_fab);
}

void imp1_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &imp1_fab);
}

void eq1_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &eq1_fab);
}

void imp2_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &imp2_fab);
}

void eq2_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &eq2_fab);
}

void imp3_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &imp3_fab);
}

void eq3_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &eq3_fab);
}

void imp4_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &imp4_fab);
}

void eq4_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &eq4_fab);
}

void imp5_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &imp5_fab);
}

void eq5_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &eq5_fab);
}


void lc_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &lc_fab);
}

void uc_fabn(FABN *f_acc, FABN *f_op) {
	forall_bin_fabn(f_acc, f_op, &uc_fab);
}

void pt_fabn(FABN *f) {
	pr_fabn(f);
	tex_fabn(f);
}

void ptb_fabn(FABN *f) {
	pr_fabn(f);
	tex_fabn(f);
	beran_fabn(f);
}

void tex_fabn(FABN *f) {
	int i;

	if (!is_bool_fabn(f)) {
		for (i=0;i<exp2(f->n);i++) {
			tex_fab(f->fabn[i]);
			if (i!=exp2(f->n)-1) printf(", ");
			else printf("\n");

			if ((i>0) && !((i+1)%6)) printf("\n");
		}
	}
}

void beran_fabn(FABN *f) {
	int i,k = exp2(f->n);

	if (!is_bool_fabn(f)) {
		if (f->n)
			printf("Beran's numbers of components: ");
		else
			printf("Beran's number of the element: ");

		for (i=0;i<k;i++) {
			printf("%3d", MO2_to_Ber(f->fabn[i]));
			if (i!=k-1) printf(", ");
			else printf("\n");
			if (i==8) printf("\n ");
			if (i>8 && !((i-8)%15)) printf("\n ");
		}
	}
}

void pr_fabn(FABN *f) {
	int i=0,j,k;


	if (!is_bool_fabn(f)) {
		while (i<exp2(f->n)) {
			for (j=1;j<=5;j++) {
				for (k=0;(i+k<exp2(f->n)) && k<6 ;k++) {
					prnthrow_fab(f->fabn[i+k],j);
					printf((j==3) && (i+k!=exp2(f->n)-1)?" , ":"   ");
				}
				printf("\n");
			}
			i+=6;
		}
	} else {
		printf(is_eq_fab(f->fabn[0],T)?"True":"False");
		printf("\n");
	}
}

/* Commutativity relation: prints f1 and (f1/\f2) \/ (f1/\f2') */
void C_pr_fabn(FABN *f1, FABN *f2) {
	FABN temp;
	pr_fabn(f1);
	
	create_fabn(&temp, f1->n);
	load_fabn(&temp, f1);
	C_rs_fabn(&temp, f2);
	pr_fabn(&temp);
	clear_fabn(&temp);
}

int C_i_fabn(FABN *f1, FABN *f2) {
	FABN temp;
	int result;

	create_fabn(&temp, f1->n);
	load_fabn(&temp, f1);
	C_rs_fabn(&temp, f2);
	result=is_eq_fabn(&temp,f1);	
	clear_fabn(&temp);
	return result;		
}

void C_f_fabn(FABN *f1, FABN *f2, FABN *f_acc) {
	addapt_bool_fabn(C_i_fabn(f1,f2), f_acc);
}

/* Dual Commutativity relation(from Beran): prints f1 and (f1\/f2)/\(f1\/f2') */
void D_pr_fabn(FABN *f1, FABN *f2) {
	FABN temp;
	pr_fabn(f1);
	
	create_fabn(&temp, f1->n);
	load_fabn(&temp, f1);
	D_rs_fabn(&temp, f2);
	pr_fabn(&temp);
	clear_fabn(&temp);
}

int D_i_fabn(FABN *f1, FABN *f2) {
	FABN temp;
	int result;

	create_fabn(&temp, f1->n);
	load_fabn(&temp, f1);
	D_rs_fabn(&temp, f2);
	result=is_eq_fabn(&temp,f1);	
	clear_fabn(&temp);
	return result;		
}

void D_f_fabn(FABN *f1, FABN *f2, FABN *f_acc) {
	addapt_bool_fabn(D_i_fabn(f1,f2), f_acc);
}

/* Inequality relations */

/* Equality */
int is_eq_fabn(FABN *f1, FABN *f2) {
	int i;
	for (i=0;i<exp2(f1->n);i++) 
		if (!is_eq_fab(f1->fabn[i], f2->fabn[i])) return 0;
	return 1;
}

void is_eq_f_fabn(FABN *f1, FABN *f2, FABN *f_acc) {
	addapt_bool_fabn(is_eq_fabn(f1,f2),f_acc);
}

int is_eq_inter_fabn(FABN *f1, FABN *f2, FABN *interpretation) {
	FABN temp1,temp2;
	int result;

	create_fabn(&temp1, f1->n);
	create_fabn(&temp2, f1->n);

	load_fabn(&temp1, f1);
	m_fabn(&temp1,interpretation);

	load_fabn(&temp2, f2);
	m_fabn(&temp2, interpretation);
	
	result=is_eq_fabn(&temp1,&temp2);

	clear_fabn(&temp1);
	clear_fabn(&temp2);

	return result;
}

void is_eq_inter_f_fabn(FABN *f1, FABN *f2, FABN *interpretation, FABN* f_acc) {
	addapt_bool_fabn(is_eq_inter_fabn(f1,f2,interpretation),f_acc);	
}

/* less than or equal */
int is_leq_fabn(FABN *f1, FABN *f2) {
	int i;
	for (i=0;i<exp2(f1->n);i++) 
		if (!is_leq_fab(f1->fabn[i], f2->fabn[i])) return 0;
	return 1;
}

void is_leq_f_fabn(FABN *f1, FABN *f2, FABN *f_acc) {
	addapt_bool_fabn(is_leq_fabn(f1,f2),f_acc);
}

int is_leq_inter_fabn(FABN *f1, FABN *f2, FABN *interpretation) {
	FABN temp1,temp2;
	int result;

	create_fabn(&temp1, f1->n);
	create_fabn(&temp2, f1->n);

	load_fabn(&temp1, f1);
	m_fabn(&temp1,interpretation);

	load_fabn(&temp2, f2);
	m_fabn(&temp2, interpretation);
	
	result=is_leq_fabn(&temp1,&temp2);

	clear_fabn(&temp1);
	clear_fabn(&temp2);

	return result;
}

void is_leq_inter_f_fabn(FABN *f1, FABN *f2, FABN *interpretation, FABN* f_acc) {
	addapt_bool_fabn(is_leq_inter_fabn(f1,f2,interpretation),f_acc);	
}


/* greater than or equal */
int is_geq_fabn(FABN *f1, FABN *f2) {
	int i;
	for (i=0;i<exp2(f1->n);i++) 
		if (!is_geq_fab(f1->fabn[i], f2->fabn[i])) return 0;
	return 1;
}

void is_geq_f_fabn(FABN *f1, FABN *f2, FABN *f_acc) {
	addapt_bool_fabn(is_geq_fabn(f1,f2),f_acc);
}

int is_geq_inter_fabn(FABN *f1, FABN *f2, FABN *interpretation) {
	FABN temp1,temp2;
	int result;

	create_fabn(&temp1, f1->n);
	create_fabn(&temp2, f1->n);

	load_fabn(&temp1, f1);
	m_fabn(&temp1,interpretation);

	load_fabn(&temp2, f2);
	m_fabn(&temp2, interpretation);
	
	result=is_geq_fabn(&temp1,&temp2);

	clear_fabn(&temp1);
	clear_fabn(&temp2);

	return result;
}

void is_geq_inter_f_fabn(FABN *f1, FABN *f2, FABN *interpretation, FABN* f_acc) {
	addapt_bool_fabn(is_geq_inter_fabn(f1,f2,interpretation),f_acc);	
}

/* Interpretations */

void create_interpretation_eq_fabn(FABN *f1, FABN *f2, FABN *interpretation) {
	FAB temp;
	int i;

	for (i=0;i<exp2(f1->n);i++) {
		create_interpretation_eq_fab(f1->fabn[i],f2->fabn[i],&temp);
		interpretation->fabn[i]=temp;
	}
}

void create_interpretation_leq_fabn(FABN *f1, FABN *f2, FABN *interpretation) {
	FAB temp;
	int i;

	for (i=0;i<exp2(f1->n);i++) {
		create_interpretation_leq_fab(f1->fabn[i],f2->fabn[i],&temp);
		interpretation->fabn[i]=temp;
	}

}

void create_interpretation_geq_fabn(FABN *f1, FABN *f2, FABN *interpretation) {
	FAB temp;
	int i;

	for (i=0;i<exp2(f1->n);i++) {
		create_interpretation_geq_fab(f1->fabn[i],f2->fabn[i],&temp);
		interpretation->fabn[i]=temp;
	}

}

void create_interpretation_C_fabn(FABN *f1, FABN *f2, FABN *interpretation) {
	FAB temp;
	int i;

	for (i=0;i<exp2(f1->n);i++) {
		create_interpretation_eq_fab(f1->fabn[i],C_rs_fab(f1->fabn[i],f2->fabn[i]),&temp);
		interpretation->fabn[i]=temp;
	}

} 

void create_interpretation_D_fabn(FABN *f1, FABN *f2, FABN *interpretation) {
	FAB temp;
	int i;

	for (i=0;i<exp2(f1->n);i++) {
		create_interpretation_eq_fab(f1->fabn[i],D_rs_fab(f1->fabn[i],f2->fabn[i]),&temp);
		interpretation->fabn[i]=temp;
	}

} 

/* Conditional operations */

void cond_eq_fabn(FABN *f_acc, FABN *f_op) {
	FABN temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_eq_fabn(&temp,f_op,f_acc);

	clear_fabn(&temp);
}

void cond_eq_fabn_rel(FABN *f_acc, FABN *f_op, int eval_as_rel) {
	FABN temp, I_temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_eq_fabn(&temp,f_op,f_acc);
	clear_fabn(&temp);
	if (eval_as_rel) {
		create_fabn(&I_temp, f_acc->n);
		I_fabn(&I_temp);
		addapt_bool_fabn(is_eq_fabn(f_acc, &I_temp), f_acc);
		clear_fabn(&I_temp);
	}

}

void cond_leq_fabn(FABN *f_acc, FABN *f_op) {
	FABN temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_leq_fabn(&temp,f_op,f_acc);

	clear_fabn(&temp);
}

void cond_leq_fabn_rel(FABN *f_acc, FABN *f_op, int eval_as_rel) {
	FABN temp, I_temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_leq_fabn(&temp,f_op,f_acc);
	clear_fabn(&temp);
	if (eval_as_rel) {
		create_fabn(&I_temp, f_acc->n);
		I_fabn(&I_temp);
		addapt_bool_fabn(is_eq_fabn(f_acc, &I_temp), f_acc);
		clear_fabn(&I_temp);
	}
}

void cond_geq_fabn(FABN *f_acc, FABN *f_op) {
	FABN temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_geq_fabn(&temp,f_op,f_acc);

	clear_fabn(&temp);
}

void cond_geq_fabn_rel(FABN *f_acc, FABN *f_op, int eval_as_rel) {
	FABN temp, I_temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_geq_fabn(&temp,f_op,f_acc);
	clear_fabn(&temp);
	if (eval_as_rel) {
		create_fabn(&I_temp, f_acc->n);
		I_fabn(&I_temp);
		addapt_bool_fabn(is_eq_fabn(f_acc, &I_temp), f_acc);
		clear_fabn(&I_temp);
	}
}

void cond_C_fabn(FABN *f_acc, FABN *f_op) {
	FABN temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_C_fabn(&temp,f_op,f_acc);

	clear_fabn(&temp);
}

void cond_C_fabn_rel(FABN *f_acc, FABN *f_op, int eval_as_rel) {
	FABN temp, I_temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_C_fabn(&temp,f_op,f_acc);
	clear_fabn(&temp);
	if (eval_as_rel) {
		create_fabn(&I_temp, f_acc->n);
		I_fabn(&I_temp);
		addapt_bool_fabn(is_eq_fabn(f_acc, &I_temp), f_acc);
		clear_fabn(&I_temp);
	}
}

void cond_D_fabn(FABN *f_acc, FABN *f_op) {
	FABN temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_D_fabn(&temp,f_op,f_acc);

	clear_fabn(&temp);
}

void cond_D_fabn_rel(FABN *f_acc, FABN *f_op, int eval_as_rel) {
	FABN temp, I_temp;

	create_fabn(&temp, f_acc->n);

	load_fabn(&temp, f_acc);
	create_interpretation_D_fabn(&temp,f_op,f_acc);
	clear_fabn(&temp);
	if (eval_as_rel) {
		create_fabn(&I_temp, f_acc->n);
		I_fabn(&I_temp);
		addapt_bool_fabn(is_eq_fabn(f_acc, &I_temp), f_acc);
		clear_fabn(&I_temp);
	}
}


void cond_and_fabn(FABN *f_acc, FABN *f_op) {
	m_fabn(f_acc,f_op);
}

void cond_and_fabn_rel(FABN *f_acc, FABN *f_op, int eval_as_rel) {
	if (!eval_as_rel)
		m_fabn(f_acc,f_op);
	else {
		FABN T_temp;
		create_fabn(&T_temp, f_acc->n);
		T_fabn(&T_temp);
		
		addapt_bool_fabn(is_eq_fabn(f_acc, &T_temp) && is_eq_fabn(f_op, &T_temp), f_acc);
		clear_fabn(&T_temp);
		
	}
}

void cond_or_fabn(FABN *f_acc, FABN *f_op) {
	j_fabn(f_acc,f_op);
}

void cond_or_fabn_rel(FABN *f_acc, FABN *f_op, int eval_as_rel) {
	if (!eval_as_rel)
		j_fabn(f_acc,f_op);
	else {
		FABN T_temp;
		create_fabn(&T_temp, f_acc->n);
		T_fabn(&T_temp);
		
		addapt_bool_fabn(is_eq_fabn(f_acc, &T_temp) || is_eq_fabn(f_op, &T_temp), f_acc);
		clear_fabn(&T_temp);
		
	}
}

void cond_if_fabn(FABN *inter_facc, FABN *inter_fop) {
	addapt_bool_fabn(is_leq_fabn(inter_facc,inter_fop),inter_facc);
}

void cond_iff_fabn(FABN *inter_facc, FABN *inter_fop) {
	addapt_bool_fabn(is_eq_fabn(inter_facc,inter_fop),inter_facc);
}

extern void eval_RPN_fabn_ver3(char *rpn, FABN *result, int eval_as_rel); 
extern int create_std_values_fabn(int no_of_comm, FABN **values); 
extern void clear_std_values_fabn(int no_of_comm, FABN **values); 
extern void eval_RPN_in_fabn_ver3(char *rpn, FABN **values,FABN *result, int eval_as_rel);

void Ber_to_MO2_fabn(int ber_val, int no_of_gen, FABN *result) {
	Dynamic_str rpn_copy;
	init_string(&rpn_copy);
	append_string(&rpn_copy, (char *)beran_const_rpn[ber_val-1]);
	rpn_copy.str[0]='0'+no_of_gen;

	eval_RPN_fabn_ver3(rpn_copy.str,result, 0);
}

void Ber_to_MO2_fabn_ext(int ber_val, int no_of_gen, FABN *result, FABN *arg_a, FABN *arg_b) {
	Dynamic_str rpn_copy;
	FABN *var_values;

	init_string(&rpn_copy);
	append_string(&rpn_copy, (char *)beran_const_rpn[ber_val-1]);
	rpn_copy.str[0]='0'+no_of_gen;
	create_std_values_fabn(no_of_gen, &var_values);
	load_fabn(&var_values[0], arg_a);
	load_fabn(&var_values[1], arg_b);
	eval_RPN_in_fabn_ver3(rpn_copy.str,&var_values, result, 0);
	clear_std_values_fabn(no_of_gen, &var_values);
	release_string(&rpn_copy);
}
