#include <stdio.h>
#include <stdlib.h>
#include "fabn.h"
#include "fabn_dynstack.h"


void init_dynstack_fabn(DYNSTACK_FABN *st) {
	
	st->bottom = (FABN_LIST *)malloc(sizeof(FABN_LIST));
	if (st->bottom == NULL) {
		fprintf(stderr, "Memory allocation error. (init_dynstack_fabn)");
		return;
	}
	st->bottom->next = NULL;
	st->top=st->bottom;
}

int is_empty_dynstack_fabn(DYNSTACK_FABN *st) {
	return st->bottom==st->top;
}

int push_dynstack_fabn(DYNSTACK_FABN *st, FABN *element) {
	
	FABN_LIST *pom;

	pom = (FABN_LIST *)malloc(sizeof(FABN_LIST));
	if (pom == NULL) {
		fprintf(stderr, "Not enough memory: Could not allocate memory for struct FABN_LIST.\n");
		return -1;
	}	

	pom->next=st->top;
	create_fabn(&pom->element,element->n);
	load_fabn(&pom->element,element);
	st->top=pom;

	return 0;
}

int pop_dynstack_fabn(DYNSTACK_FABN *st, FABN *element) {
	FABN_LIST *pom;

	if (is_empty_dynstack_fabn(st)) {
		fprintf(stderr, "Stack is empty. (Error in the epression.)\n");
		return -1;
	}

/*	pom=st->top; */ /* ????  */
	load_fabn(element,&st->top->element);
	clear_fabn(&st->top->element);
	
	pom=st->top->next;
	free((void *)st->top);
	st->top=pom;

	return 0;
}

void flush_dynstack_fabn(DYNSTACK_FABN *st) {
	FABN elem;
	int is_created=0;

	while (!is_empty_dynstack_fabn(st)) {
		if (!is_created) {
			create_fabn(&elem,st->top->element.n);
			is_created=1;
			
		}
		pop_dynstack_fabn(st, &elem);
	}

	if (is_created)
		clear_fabn(&elem);

	free((void *)st->bottom);
}

/* NEW_DYNSTACK */

void init_new_dynstack_fabn(NEW_DYNSTACK_FABN *st) {

	st->bottom = (NEW_FABN_LIST *)malloc(sizeof(NEW_FABN_LIST));
	if (st->bottom == NULL) {
		fprintf(stderr, "Memory allocation error (init_new_dynstack_fabn)");
		return;
	}
	
	st->bottom->next = NULL;
	st->top = st->bottom;
}

int is_empty_new_dynstack_fabn(NEW_DYNSTACK_FABN *st) {
	return st->bottom==st->top;
}

int push_int_new_dynstack_fabn(NEW_DYNSTACK_FABN *st, int i_el) {
	t_stack_element_fabn el;
	el.i_element=i_el;
#ifdef DEBUG
	fprintf(stderr, "Pushing INT...\n");
#endif
	return push_new_dynstack_fabn(st, &el, INT_TYPE);
}

int push_fabn_new_dynstack_fabn(NEW_DYNSTACK_FABN *st, FABN *fn_el) {
	t_stack_element_fabn el;
	el.fn_element.n=fn_el->n;
	el.fn_element.fabn = fn_el->fabn;
#ifdef DEBUG
	fprintf(stderr, "Pushing FABN...\n");
#endif
	return push_new_dynstack_fabn(st, &el, FABN_TYPE);
}

int push_new_dynstack_fabn(NEW_DYNSTACK_FABN *st, t_stack_element_fabn *el, t_element_type_fabn type) {
	NEW_FABN_LIST *pom;

	pom = (NEW_FABN_LIST *)malloc(sizeof(NEW_FABN_LIST));
	if (pom == NULL) {
		fprintf(stderr, "Not enough memory: Could not allocate memory for struct NEW_FABN_LIST.\n");
		return -1;
	}	

	switch (type) {
		case INT_TYPE:
			pom->el.i_element=el->i_element;
			break;
		case FABN_TYPE:
			if (create_fabn(&pom->el.fn_element,el->fn_element.n)<0) {
				fprintf(stderr, "Error: Cannot allocate memory for FABN. (push_new_dynstack_fabn)\n");
				free ((void *)pom);
				return -1;
			};
			load_fabn(&pom->el.fn_element,&el->fn_element);
			break;
		default:
			fprintf(stderr, "Error: Unknown stack type!\n");
			free((void *)pom);
			return -1;
			break;
	}
	pom->element_type=type;
	pom->next=st->top;
	st->top=pom;

	return 0;
}

int pop_int_new_dynstack_fabn(NEW_DYNSTACK_FABN *st, int *i_element) {
	NEW_FABN_LIST *pom;

	if (is_empty_new_dynstack_fabn(st)) {
		fprintf(stderr, "Stack is empty. (Error in the epression.)\n");
		return -1;
	}

	if (st->top->element_type != INT_TYPE) {
		fprintf(stderr, "Error: Element in stack is not INT element.\n");
		return -1;
	}
	*i_element = st->top->el.i_element;

	pom=st->top->next;
	free((void *)st->top);
	st->top=pom;
#ifdef DEBUG
	fprintf(stderr, "Popping INT...\n");
#endif 
	return 0;
}

/* Memory for fn_element is already allocated */
int pop_fabn_new_dynstack_fabn(NEW_DYNSTACK_FABN *st, FABN *fn_element) {
	NEW_FABN_LIST *pom;

	if (is_empty_new_dynstack_fabn(st)) {
		fprintf(stderr, "Stack is empty. (Error in the epression.)\n");
		return -1;
	}

	if (st->top->element_type != FABN_TYPE) {
		fprintf(stderr, "Error: Element in stack is not FABN element.\n");
		return -1;
	}
	if (st->top->el.fn_element.n != fn_element->n) {
		fprintf(stderr, "Error: The size of desired FABN element (%d) and stack FABN element (%d) does not match.\n", st->top->el.fn_element.n, fn_element->n);
		return -1;
	}
	load_fabn(fn_element, &st->top->el.fn_element);

	pom=st->top->next;
	clear_fabn(&st->top->el.fn_element);
	free((void *)st->top);
	st->top=pom;
#ifdef DEBUG
	fprintf(stderr, "Popping FABN...\n");
#endif
	return 0;
}

void flush_new_dynstack_fabn(NEW_DYNSTACK_FABN *st) {
	NEW_FABN_LIST *pom;

	while (!is_empty_new_dynstack_fabn(st)) {
		if (st->top->element_type == FABN_TYPE)
			clear_fabn(&st->top->el.fn_element);
		pom = st->top->next;
		free((void *)st->top);
		st->top = pom;
	}
	
	free((void *)st->bottom);
}
