#include <stdio.h>
#include <string.h>
#include "fabx_rpn.h"
#include "fabx_parser.h"
#include "op_dynstack.h"

void display_error(char *parsed, char *pom_pom) {
	char *pom;

	fprintf(stderr,"%s\n",parsed);
	for (pom=parsed;pom<pom_pom;pom++) fprintf(stderr," ");
	fprintf(stderr,"^\n");
}

void append_int_to_rpn(Dynamic_str *rpn_str, int number) {
	char num[5]={'\0', '\0', '\0', '\0', '\0'};
	sprintf(num, "%02d", number);
	append_string(rpn_str, "I");
	append_string(rpn_str, num);
	append_string(rpn_str, " ");
}
void display_error_with_limit(char *parsed, char *pom_pom, int limit_size) {
#if 0
	char *pom;

	fprintf(stderr,"%s\n",parsed);
	for (pom=parsed;pom<pom_pom;pom++) fprintf(stderr," ");
	fprintf(stderr,"^\n");
#endif
	/* TODO */
}


int to_RPN_ver5_autodetect(char *parsed, int *no_of_gen, Dynamic_str *output) {
	char *pom=parsed, *pom_pom;
	char gen_str[2]={'\0','\0'};
	int status, pom_val;
	OPTYPE current,op_pom, 
	last_processed=OP_NONE, last_nonws_processed=OP_NONE;
	DYNSTACK_OP stack;
	
	*no_of_gen=0;
	init_dynstack_op(&stack);

	gen_str[0]='0';

	append_string(output, gen_str);
	while ((status = parse_next_element_fabx_ver5(pom, &pom_pom, &current))==PARSE_OK) {
		switch(get_opkind(current)) {
		default: break;
		case ARG_COMM:
			if (get_opkind(last_processed)==DELIM) {
				last_processed=last_nonws_processed;
				last_nonws_processed=OP_NONE;
			}
			if (get_opkind(last_processed) == ARG      ||
			    get_opkind(last_processed) == ARG_COMM ||
			    get_opkind(last_processed) == OP1PO) {
				fprintf(stderr, "Error: Missing an operation for the operand.\n");
				flush_dynstack_op(&stack);
				display_error(parsed, pom);
				return -1;
			}
			
			if ((pom_val=parsed_elements[find_optype_idx(current)].inv_priority)>*no_of_gen) {
#ifdef DEBUG
				fprintf(stderr, "Setting new no_of_gen value: %d", pom_val);
#endif
				*no_of_gen=pom_val;
			}

			append_string(output, 
			parsed_elements[find_optype_idx(current)].rpn_string);

			last_processed=current;
			last_nonws_processed=current;
			break;
		case ARG:
			if (get_opkind(last_processed)==DELIM) {
				last_processed=last_nonws_processed;
				last_nonws_processed=OP_NONE;
			}
			if (get_opkind(last_processed) == ARG      ||
			    get_opkind(last_processed) == ARG_COMM ||
			    get_opkind(last_processed) == OP1PO) {
				fprintf(stderr, "Error: <ARG> Missing operation for the operand.\n");
				flush_dynstack_op(&stack);
				display_error(parsed, pom);
				return -1;
			}
			
			append_string(output, 
			parsed_elements[find_optype_idx(current)].rpn_string);
			last_processed=current;
			last_nonws_processed=current;
			break;
		case OP1PO:
			if (last_processed == OP_RIGHTPAR 		|| 
			    get_opkind(last_processed) == ARG 		||
			    get_opkind(last_processed) == ARG_COMM 	||
			    get_opkind(last_processed) == OP1PO) {
				append_string(output, 
				parsed_elements[find_optype_idx(current)].rpn_string);
				last_processed=current;
				last_nonws_processed=current;
			}
			else {
				fprintf(stderr,"Syntax Error: Missing argument for unary postfix operator.\n");
				flush_dynstack_op(&stack);
				display_error(parsed, pom);
				return -1;
			}
			break;
		case PAR:
			if (current == OP_LEFTPAR) {
				if (push_dynstack_op(&stack, current)!=0) {
		 			flush_dynstack_op(&stack);
				display_error(parsed, pom);
					return -1;
				}
			} else 
			if (current == OP_RIGHTPAR) {

				if (get_opkind(last_processed)==DELIM) {
					last_processed=last_nonws_processed;
					last_nonws_processed=OP_NONE;
				}

				if (last_processed==OP_LEFTPAR) {
					fprintf(stderr, "Syntax error: Empty expression in parenthesis.\n");
					flush_dynstack_op(&stack);
					display_error(parsed, pom);
					return -1;
				}

				if (get_opkind(last_processed)==OP1PR) {
					fprintf(stderr, "Syntax error: Missing argument of unary prefix operator.\n");
					flush_dynstack_op(&stack);
				display_error(parsed, pom);
					return -1;
				}

				if (get_opkind(last_processed)==OP2IN) {
					fprintf(stderr, "Syntax error: Missing a second argument of the operation/relation.\n");
					flush_dynstack_op(&stack);
				display_error(parsed, pom);
					return -1;
				}

				do {
					if (is_empty_dynstack_op(&stack)) {
						fprintf(stderr,"Syntax Error.\n");
						flush_dynstack_op(&stack);
				display_error(parsed, pom);
						return -1;
					}
					pop_dynstack_op(&stack, &op_pom);
					if (op_pom!=OP_LEFTPAR) {
						append_string(output, 
						 parsed_elements[find_optype_idx(op_pom)].rpn_string);
					}
				} while( op_pom !=OP_LEFTPAR);
			}
			last_processed=current;
			last_nonws_processed=current;
			break;
		case OP2IN:
			if (get_opkind(last_processed)== DELIM) {
				last_processed=last_nonws_processed;
				last_nonws_processed=OP_NONE;
			}

			if (get_opkind(last_processed)!=OP1PO 	&&
			    last_processed!=OP_RIGHTPAR 	&&
			    get_opkind(last_processed)!=ARG	&&
			    get_opkind(last_processed)!=ARG_COMM &&
			    get_opkind(last_processed)!=FUN) {
				fprintf(stderr,"Error: Missing first operand for binary operation.\n");
				flush_dynstack_op(&stack);
				display_error(parsed, pom);
				return -1;
			}
			break;
		case OP1PR:
			/* Flush all with priority geq , i.e */
			/* inv priority leq */
			while (!is_empty_dynstack_op(&stack)) {
				pop_dynstack_op(&stack, &op_pom);
				if (parsed_elements[find_optype_idx(op_pom)].inv_priority<= parsed_elements[find_optype_idx(current)].inv_priority /*&& !is_rel(op_pom)*/) {
					append_string(output, 
					parsed_elements[find_optype_idx(op_pom)].rpn_string);
				}
				else {
					push_dynstack_op(&stack, op_pom);
					break;
				}
			}
			/* push into stack */
			if (push_dynstack_op(&stack, current)!=0) {
				return -1;
			}
			last_processed=current;
			last_nonws_processed=current;
			break;
		case FUN: 
			switch (current) {
				case BER1:
					pom=pom_pom;
					{
						char *pom_pom;
						int value;
					if (*pom!='(') {
						fprintf(stderr, "Error: Missing openning parenthesis in function call.\n");
						return -1;
					}
					pom++;
					if (parse_next_element_int(pom, &pom_pom, &value)!=PARSE_OK) {
						fprintf(stderr, "Error: Missing integer argument.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						return -1;
					}
					pom=pom_pom;

					if (value<1 || value >96) {
						fprintf(stderr, "Error: Integer value out of range.\n Only values 1<=x<=96 are expected.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						return -1;
					}
					

					while (*pom==' ' || *pom=='\t') pom++;
					if (*pom!=')') {
						fprintf(stderr, "Error: Missing closing parenthesis in calling B1 function.\n");
						flush_dynstack_op(&stack);
						display_error(parsed,pom);
						return -1;
					}
					pom++;
					{
						char num[5];
						int i;
						for (i=0;i<5;i++) num[i]='\0';
						sprintf(num, "%02d", value);
						append_string(output, "I");
						append_string(output, num);
						append_string(output, " ");
					}
					append_string(output, 
					parsed_elements[find_optype_idx(current)].rpn_string);
					}
					pom_pom=pom;	
					last_processed=current;
					last_nonws_processed=current;
					break;
#if 0
				case BER3: {
						   int status, value;
						
					if (*pom!='(') {
						fprintf(stderr, "Error: Missing parenthesis in function call.\n");
						return -1;
					}
					pom++;
					if ((status=parse_next_element_int(pom, pom_pom, &value))!=PARSE_OK) {
						/*ERROR*/
						return -1;
					}
					else {
						Dynamic_str arg2, arg3;
						int nogen2, nogen3;
					
						/* init_dyn_str */	
						to_RPN_ver5_autodetect(pom, &nogen2, &arg2);

						to_RPN_ver5_autodetect(pom, &nogen3, &arg3); 
					
					}
					   }
					break;
#endif
				default:
					break;
			}
			/* push_dynstack(B?)
			 * if 
			 *
			 *
			 */
			break;
		case ARG_DELIM:
			break;
		} /* switch */
		pom=pom_pom;
	} /* while */
	if (status==-1) {
		fprintf(stderr, "Syntax error: Unknown or incomplete expression.\n");
		flush_dynstack_op(&stack); 
		display_error(parsed, pom);
		return -1;
	}
	if (get_opkind(last_processed)==DELIM) {
		last_processed=last_nonws_processed;
		last_nonws_processed=OP_NONE;
	}

	if (get_opkind(last_processed)==OP1PR) {
		fprintf(stderr, "Syntax error: Missing argument of unary prefix operator.\n");
		flush_dynstack_op(&stack);
		display_error(parsed, pom);
		return -1;
	}

	if (get_opkind(last_processed)==OP2IN) {
		fprintf(stderr, "Syntax error: Missing a second argument of the operation/relation.\n");
		flush_dynstack_op(&stack);
		display_error(parsed, pom);
		return -1;
	}

#ifdef DEBUG
	fprintf(stderr, "\n HERE....!!\n");
#endif
	while (!is_empty_dynstack_op(&stack)) {
		pop_dynstack_op(&stack, &current);
		if (current!=OP_LEFTPAR) {
			append_string(output,
			 parsed_elements[find_optype_idx(current)].rpn_string);
		}
		else {
			fprintf(stderr, "Syntax error: Missing closing parenthesis.\n");
			flush_dynstack_op(&stack); 
		display_error(parsed, pom);
			return -1;
		}
	}
	flush_dynstack_op(&stack);
	
	output->str[0]='0'+*no_of_gen;
	return 0;
}

int to_RPN_ver6_autodetect(char *parsed, char **pom_pom, int *no_of_gen, Dynamic_str *output, t_parser_state state) {
	char *pom=parsed;
	char gen_str[2]={'\0','\0'};
	int status, pom_val;
	OPTYPE current,op_pom, 
	last_processed=OP_NONE, last_nonws_processed=OP_NONE;
	DYNSTACK_OP stack;
	
	*no_of_gen=0;
	init_dynstack_op(&stack);

	gen_str[0]='0';

#ifdef DEBUG
	fprintf(stderr, "PARSED: %s", pom);
#endif
	/* add number of generators */
	append_string(output, gen_str);
	while ((status = parse_next_element_fabx_ver5(pom, pom_pom, &current))==PARSE_OK) {
		switch(get_opkind(current)) {
		default: /*printf("\n---HERE---\n");*/ break;
		case ARG_COMM:
#ifdef DEBUG
			fprintf(stderr, "case ARG_COMM:\n");
#endif
			if (get_opkind(last_processed)==DELIM) {
				last_processed=last_nonws_processed;
				last_nonws_processed=OP_NONE;
			}
			if (get_opkind(last_processed) == ARG      ||
			    get_opkind(last_processed) == ARG_COMM ||
			    get_opkind(last_processed) == OP1PO) {
				fprintf(stderr, "Error: Missing an operation for the operand.\n");
				flush_dynstack_op(&stack);
				display_error(parsed, pom);
				return -1;
			}
			
			if ((pom_val=parsed_elements[find_optype_idx(current)].inv_priority)>*no_of_gen) {
#ifdef DEBUG
				fprintf(stderr, "Setting new no_of_gen value: %d", pom_val);
#endif
				*no_of_gen=pom_val;
			}

			append_string(output, 
			parsed_elements[find_optype_idx(current)].rpn_string);

			last_processed=current;
			last_nonws_processed=current;
			break;
		case ARG:
#ifdef DEBUG
			fprintf(stderr, "case ARG:\n");
#endif
			if (get_opkind(last_processed)==DELIM) {
				last_processed=last_nonws_processed;
				last_nonws_processed=OP_NONE;
			}
			if (get_opkind(last_processed) == ARG      ||
			    get_opkind(last_processed) == ARG_COMM ||
			    get_opkind(last_processed) == OP1PO) {
				fprintf(stderr, "Error: <ARG> Missing operation for the operand.\n");
				flush_dynstack_op(&stack);
				display_error(parsed, pom);
				return -1;
			}
			
			append_string(output, 
			parsed_elements[find_optype_idx(current)].rpn_string);
			last_processed=current;
			last_nonws_processed=current;
			break;
		case OP1PO:
#ifdef DEBUG
			fprintf(stderr, "case OP1PO:\n");
#endif
			if (last_processed == OP_RIGHTPAR 		|| 
			    get_opkind(last_processed) == ARG 		||
			    get_opkind(last_processed) == ARG_COMM 	||
			    get_opkind(last_processed) == OP1PO		||
			    get_opkind(last_processed) == FUN) {
				append_string(output, 
				parsed_elements[find_optype_idx(current)].rpn_string);
				last_processed=current;
				last_nonws_processed=current;
			}
			else {
				fprintf(stderr,"Syntax Error: Missing argument for unary postfix operator.\n");
				flush_dynstack_op(&stack);
				display_error(parsed, pom);
				return -1;
			}
			break;
		case PAR:
#ifdef DEBUG
			fprintf(stderr, "case PAR:\n");
#endif
			if (current == OP_LEFTPAR) {
				if (push_dynstack_op(&stack, current)!=0) {
		 			flush_dynstack_op(&stack);
				display_error(parsed, pom);
					return -1;
				}
			} else 
			if (current == OP_RIGHTPAR) {

				if (get_opkind(last_processed)==DELIM) {
					last_processed=last_nonws_processed;
					last_nonws_processed=OP_NONE;
				}

				if (last_processed==OP_LEFTPAR) {
					fprintf(stderr, "Syntax error: Empty expression in parenthesis.\n");
					flush_dynstack_op(&stack);
					display_error(parsed, pom);
					return -1;
				}

				if (get_opkind(last_processed)==OP1PR) {
					fprintf(stderr, "Syntax error: Missing argument of unary prefix operator.\n");
					flush_dynstack_op(&stack);
				display_error(parsed, pom);
					return -1;
				}

				if (get_opkind(last_processed)==OP2IN) {
					fprintf(stderr, "Syntax error: Missing a second argument of the operation/relation.\n");
					flush_dynstack_op(&stack);
				display_error(parsed, pom);
					return -1;
				}

				switch (state) {
					case NORMAL_MODE:	
#ifdef DEBUG
			fprintf(stderr, "case PAR ... NORMAL_MODE:\n");
#endif
						do {
							if (is_empty_dynstack_op(&stack)) {
								fprintf(stderr,"Syntax Error.\n");
								flush_dynstack_op(&stack);
						display_error(parsed, pom);
								return -1;
							}
							pop_dynstack_op(&stack, &op_pom);
							if (op_pom!=OP_LEFTPAR) {
								append_string(output, 
								 parsed_elements[find_optype_idx(op_pom)].rpn_string);
							}
						} while( op_pom !=OP_LEFTPAR);
#ifdef DEBUG
	fprintf(stderr, "\nHERE\n");
#endif
						break;
					case LASTARG_MODE:
#ifdef DEBUG
						fprintf(stderr, "case PAR ... LASTARG_MODE:\n");
#endif
						{
							int is_empty=0;
							do {
								/*
								if (is_empty_dynstack_op(&stack)) {
									fprintf(stderr,"Syntax Error.\n");
							display_error(parsed, pom);
									return -1;
								}*/
								is_empty=is_empty_dynstack_op(&stack);
				
								
								if (!is_empty) {
									pop_dynstack_op(&stack, &op_pom);
									if (op_pom!=OP_LEFTPAR) {
										append_string(output, 
										 parsed_elements[find_optype_idx(op_pom)].rpn_string);
									}
								}
							} while( !is_empty && op_pom !=OP_LEFTPAR);
							if (is_empty) {
								*pom_pom=*pom_pom-1;
								flush_dynstack_op(&stack);
								return PARSE_OK;
							} /* closing parenthesis belongs to another expression*/
						}
						break;
				}
			}
			last_processed=current;
			last_nonws_processed=current;
			break;	
		case OP2IN:
#ifdef DEBUG
			fprintf(stderr, "case OP2IN:\n");
#endif
			if (get_opkind(last_processed)== DELIM) {
				last_processed=last_nonws_processed;
				last_nonws_processed=OP_NONE;
			}

			if (get_opkind(last_processed)!=OP1PO 	&&
			    last_processed!=OP_RIGHTPAR 	&&
			    get_opkind(last_processed)!=ARG	&&
			    get_opkind(last_processed)!=ARG_COMM &&
			    get_opkind(last_processed)!=FUN) {
				fprintf(stderr,"Error: Missing first operand for binary operation.\n");
				flush_dynstack_op(&stack);
				display_error(parsed, pom);
				return -1;
			}
		case OP1PR:
#ifdef DEBUG
			fprintf(stderr, "case OP1PR:\n");
#endif
			/* Flush all with priority geq , i.e */
			/* inv priority leq */
			while (!is_empty_dynstack_op(&stack)) {
				pop_dynstack_op(&stack, &op_pom);
				if (parsed_elements[find_optype_idx(op_pom)].inv_priority<= parsed_elements[find_optype_idx(current)].inv_priority /*&& !is_rel(op_pom)*/) {
					append_string(output, 
					parsed_elements[find_optype_idx(op_pom)].rpn_string);
				}
				else {
					push_dynstack_op(&stack, op_pom);
					break;
				}
			}
			/* push into stack */
			if (push_dynstack_op(&stack, current)!=0) {
				return -1;
			}
			last_processed=current;
			last_nonws_processed=current;
			break;
		case FUN: 
#ifdef DEBUG
			fprintf(stderr, "case FUN:\n");
#endif
			switch (current) {
				case BER1:
#ifdef DEBUG
			fprintf(stderr, "case FUN -> BER1:\n");
#endif
					pom=*pom_pom;
					{
						char *local_pom_pom;
						int value;
					if (*pom!='(') {
						fprintf(stderr, "Error: Missing openning parenthesis in function call.\n");
						return -1;
					}
					pom++;
					if (parse_next_element_int(pom, &local_pom_pom, &value)!=PARSE_OK) {
						fprintf(stderr, "Error: Missing integer argument.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						return -1;
					}
					pom=local_pom_pom;

					if (value<1 || value >96) {
						fprintf(stderr, "Error: Integer value out of range.\n Only values 1<=x<=96 are expected.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						return -1;
					}
					

					while (*pom==' ' || *pom=='\t') pom++;
					if (*pom!=')') {
						fprintf(stderr, "Error: Missing closing parenthesis in calling B1 function.\n");
						flush_dynstack_op(&stack);
						display_error(parsed,pom);
						return -1;
					}
					pom++;
					append_int_to_rpn(output, value);
					/* Output rpn for B1 operand */
					append_string(output, 
					parsed_elements[find_optype_idx(current)].rpn_string);
					}
					*pom_pom=pom;	
					last_processed=current;
					last_nonws_processed=current;
					break;
				/*end   case BER1*/
				/*begin case BER3*/
				case BER3: 
#ifdef DEBUG
			fprintf(stderr, "case FUN -> BER3:\n");
#endif
					pom = *pom_pom;
					{
					/* local variables*/   
				   	int value;
					char *local_pom_pom;
					Dynamic_str expr_par2, expr_par3;
					int no_gen2, no_gen3;
							
					if (*pom!='(') {
						fprintf(stderr, "Error: Missing openning parenthesis in function call.\n");
						return -1;
					}
					pom++;
					if (parse_next_element_int(pom, &local_pom_pom, &value)!=PARSE_OK) {
						fprintf(stderr, "Error: Missing integer argument.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						return -1;
					}
					pom=local_pom_pom;

					if (value<1 || value >96) {
						fprintf(stderr, "Error: Integer value out of range.\n Only values 1<=x<=96 are expected.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						return -1;
					}
					
					while (*pom==' ' || *pom=='\t') pom++;
					if (*pom!=',') {
						fprintf(stderr, "Error: Missing argument delimiter between first and second argument in call of B3 function.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						return -1;
					
					}	
					pom++;
					while (*pom==' ' || *pom=='\t') pom++;
					init_string(&expr_par2);
	#ifdef DEBUG
		fprintf(stderr, "\nLocal parsing arg2 ...\n");
	#endif 
					if (to_RPN_ver6_autodetect(pom, &local_pom_pom, &no_gen2, &expr_par2,NORMAL_MODE)!=PARSE_OK) {
						fprintf(stderr, "Error: Problem with the second argument of B3 function.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						release_string(&expr_par2);
						return -1;
					
					}
	#ifdef DEBUG
		fprintf(stderr, "\nEnd of local parsing arg2 ...\n");
	#endif 
					pom=local_pom_pom;
#ifdef DEBUG
					printf("DEBUG:\n");
						display_error(parsed, pom);
					printf("DEBUG:\n");
#endif
					
					while (*pom==' ' || *pom=='\t') pom++;
					if (*pom!=',') {
						fprintf(stderr, "Error: Missing argument delimiter between second and third argument in call of B3 function.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						release_string(&expr_par2);
						return -1;
					}
					pom++;
					
					init_string(&expr_par3);
	#ifdef DEBUG
		fprintf(stderr, "\nLocal parsing arg3 ...\n");
	#endif 
					if (to_RPN_ver6_autodetect(pom, &local_pom_pom, &no_gen3, &expr_par3,LASTARG_MODE)!=PARSE_OK) {
						fprintf(stderr, "Error: Problem with the second argument of B3 function.\n");
						flush_dynstack_op(&stack);
						display_error(parsed, pom);
						release_string(&expr_par2);
						release_string(&expr_par3);
						return -1;
					
					}
	#ifdef DEBUG
		fprintf(stderr, "\nEnd of local parsing arg3 ...\n");
	#endif 
					pom=local_pom_pom;

					while (*pom==' ' || *pom=='\t') pom++;
					/*pom++;*/ /* Skip trailing comma*/
					if (*pom!=')') {
						fprintf(stderr, "Error: Missing closing parenthesis in calling B3 function.\n");
						flush_dynstack_op(&stack);
						display_error(parsed,pom);
						release_string(&expr_par2);
						release_string(&expr_par3);
						return -1;
					
					}
					pom++;
#ifdef DEBUG
					printf("DEBUG2:\n");
						display_error(parsed,pom);
					printf("DEBUG2:\n");
#endif

					if (*no_of_gen < no_gen2) *no_of_gen = no_gen2;
					if (*no_of_gen < no_gen3) *no_of_gen = no_gen3;	
					append_string(output, &expr_par3.str[1]);
					append_string(output, &expr_par2.str[1]);
					release_string(&expr_par3);
					release_string(&expr_par2);
					append_int_to_rpn(output, value);
#if 0
					{
						char num[5];
						int i;
						for (i=0;i<5;i++) num[i]='\0';
						sprintf(num, "%02d", value);
						append_string(output, "I");
						append_string(output, num);
						append_string(output, " ");
					}
#endif
					/* append rpn string for B3 operand into */
					append_string(output, 
					parsed_elements[find_optype_idx(current)].rpn_string);
					}
					*pom_pom=pom;	
					last_processed=current;
					last_nonws_processed=current;
					break;
				/*end case BER3*/
				default:
					break;
			}
			/* push_dynstack(B?)
			 * if 
			 *
			 *
			 */
			break;
		case ARG_DELIM:
#ifdef DEBUG
			fprintf(stderr, "case ARG_DELIM:\n");
#endif
			while (!is_empty_dynstack_op(&stack)) {
				pop_dynstack_op(&stack, &op_pom);
				append_string(output, 
				parsed_elements[find_optype_idx(op_pom)].rpn_string);
			}
			flush_dynstack_op(&stack);
			*pom_pom=*pom_pom-1;
			return PARSE_OK;
			break;
		} /* switch */
		pom=*pom_pom;
#ifdef DEBUG
	fprintf(stderr, "rpn_tmp__: %s\n", output->str);
#endif
	} /* while */
#ifdef DEBUG
	fprintf(stderr,"UUUUUUU\n");
	fprintf(stderr, "status: %d\n", status);
#endif
	if (status==-1) {
		fprintf(stderr, "Syntax error: Unknown or incomplete expression.\n");
		flush_dynstack_op(&stack); 
		display_error(parsed, pom);
		return -1;
	}
	if (get_opkind(last_processed)==DELIM) {
		last_processed=last_nonws_processed;
		last_nonws_processed=OP_NONE;
	}

	if (get_opkind(last_processed)==OP1PR) {
		fprintf(stderr, "Syntax error: Missing argument of unary prefix operator.\n");
		flush_dynstack_op(&stack);
		display_error(parsed, pom);
		return -1;
	}

	if (get_opkind(last_processed)==OP2IN) {
		fprintf(stderr, "Syntax error: Missing a second argument of the operation/relation.\n");
		flush_dynstack_op(&stack);
		display_error(parsed, pom);
		return -1;
	}
#ifdef DEBUG
	fprintf(stderr,"VVVVVVV\n");
#endif

	while (!is_empty_dynstack_op(&stack)) {
#ifdef DEBUG
	fprintf(stderr,"EEEEEEE\n");
	dump_dynstack_op(&stack);
	fprintf(stderr,"----->\n");
#endif
		pop_dynstack_op(&stack, &current);
#ifdef DEBUG
	fprintf(stderr,"FFFFFFF\n");
#endif
		if (current!=OP_LEFTPAR) {
			append_string(output,
			 parsed_elements[find_optype_idx(current)].rpn_string);
		}
		else {
			fprintf(stderr, "Syntax error: Missing closing parenthesis.\n");
			flush_dynstack_op(&stack); 
		display_error(parsed, pom);
			return -1;
		}
	}
	
	flush_dynstack_op(&stack);
	output->str[0]='0'+*no_of_gen;
	return 0;
}
