#include <stdio.h>
#include <stdlib.h>
typedef struct {
	int bit_size;
	unsigned char *array;
} bitField;

int create_bitfield(bitField *fld, int size) {
	fld->array=(unsigned char *)malloc( size / 8 + (size % 8?1:0));

	if (fld->array==NULL) {
		return -1;
	}
	fld->bit_size=size;
	return 0;
}

void clear_bitfield(bitField *fld) {
	if (!fld) return;
	if (!fld->array) return;
	
	free((void *)fld->array);
	fld->array=0;
	fld->bit_size=0;
}


int get_ith_bit(bitField *fld, int bitno) {
	return (fld->array[bitno /8 + (bitno % 8?1:0)] & (1<< bitno %8)) >> (bitno % 8);
}

void set_ith_bit(bitField *fld, int bitno, int value) {
	if (!value)
		fld->array[bitno /8 +(bitno % 8?1:0)] &= ~(1<<bitno % 8); /*clear */
	else 
		fld->array[bitno /8 +(bitno % 8?1:0)] |= (1<<bitno % 8); /*set*/
	
}

void zero_bitfield(bitField *fld) {
	int i;
	for (i=0;i<fld->bit_size;i++)
		set_ith_bit(fld, i, 0);
}

int isLast_bitfield(bitField *fld) {
	int i;

	for (i=0; i<fld->bit_size;i++) 
		if (!get_ith_bit(fld, i)) return 0;
	return 1;
}

void next_bitfield(bitField *fld) {
	int i;

	for (i=0; i<fld->bit_size;i++) 
		if (!get_ith_bit(fld,i)) {set_ith_bit(fld, i, 1); break;}
		else set_ith_bit(fld, i, 0);
}
void print_bitfield(bitField *fld) {
	int i;

	for (i=0;i<fld->bit_size; i++) {
		printf(get_ith_bit(fld, fld->bit_size-i-1)?"1":"0");
	}
	printf("\n");
}

int main() {
	bitField f1,f2/*,f3*/;

	create_bitfield(&f1, 100);
	zero_bitfield(&f1);
	set_ith_bit(&f1, 30, 1);
	if (get_ith_bit(&f1, 30)) printf("Set\n"); else printf("Not set.\n");
	if (get_ith_bit(&f1, 77)) printf("Set\n"); else printf("Not set.\n");
	clear_bitfield(&f1);

	create_bitfield(&f2, 4);
	zero_bitfield(&f2);
	do {
		print_bitfield(&f2);
		next_bitfield(&f2);
	}
	while (!isLast_bitfield(&f2));
	print_bitfield(&f2);
	clear_bitfield(&f2);
	
	return 0;
}
